# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

# Released under BSD 3-Clause License,
# Copyright (c) Facebook, Inc. and its affiliates.

"""Updates FSDPs _auto_wrap to enable module_kwargs and custom process_group cache."""

import functools
import warnings
from typing import Any, Callable, Dict, Set, Tuple, Union, cast

import torch
import torch.nn as nn
from torch import distributed
from torch.distributed import ProcessGroup
from torch.distributed.fsdp import (BackwardPrefetch, CPUOffload, FullyShardedDataParallel, MixedPrecision,
                                    ShardingStrategy)
from torch.distributed.fsdp._utils import _contains_batchnorm, _override_batchnorm_mixed_precision
from torch.distributed.fsdp.wrap import _or_policy, _wrap, _wrap_batchnorm_individually

from composer.core import Precision
from composer.utils import dist

__all__ = [
    'sharding_map',
    'backward_prefetch_map',
    'get_torch_dtype',
    'get_mixed_precision',
    'get_cpu_offload',
    'get_process_group',
    'MosaicFullyShardedDataParallel',
]

sharding_map = {
    'NO_SHARD': ShardingStrategy.NO_SHARD,
    'SHARD_GRAD_OP': ShardingStrategy.SHARD_GRAD_OP,
    'FULL_SHARD': ShardingStrategy.FULL_SHARD,
}

backward_prefetch_map = {
    'NONE': None,
    'BACKWARD_PRE': BackwardPrefetch.BACKWARD_PRE,
    'BACKWARD_POST': BackwardPrefetch.BACKWARD_POST,
}


def get_torch_dtype(dtype: Union[Precision, str]):
    """Convert common string representations of dtypes to torch dtypes."""
    dtype = dtype.value if isinstance(dtype, Precision) else dtype
    if dtype in ['float32', 'torch.float32', 'fp32']:
        return torch.float32
    elif dtype in ['float16', 'torch.float16', 'half', 'fp16', 'amp', 'amp_fp16']:
        return torch.float16
    elif dtype in ['bfloat16', 'bfloat', 'torch.bfloat16', 'bf16', 'amp_bf16']:
        return torch.bfloat16
    elif dtype in ['float8', 'torch.float8', 'fp8', 'amp_fp8']:
        if hasattr(torch, 'float8'):
            raise NotImplementedError('Torch has enabled float8. This should be updated to `return torch.float8`')
        else:
            warnings.warn('We use torch.bfloat16 by default for amp_fp8 as there is no fp8 datatype in PyTorch yet.')
            return torch.bfloat16
    else:
        raise ValueError(f'Not sure how to convert dtype={dtype} to a torch dtype.')


def get_mixed_precision(precision, mixed_precision='DEFAULT', keep_low_precision_grads=False):
    """Helper function for configuring mixed_precision."""
    param_dtype = None
    reduce_dtype = None
    buffer_dtype = None
    if isinstance(mixed_precision, dict):
        param_dtype = mixed_precision.get('param_dtype', None)
        if param_dtype is not None:
            param_dtype = get_torch_dtype(param_dtype)
        reduce_dtype = mixed_precision.get('reduce_dtype', None)
        if reduce_dtype is not None:
            reduce_dtype = get_torch_dtype(reduce_dtype)
        buffer_dtype = mixed_precision.get('buffer_dtype', None)
        if buffer_dtype is not None:
            buffer_dtype = get_torch_dtype(buffer_dtype)
    elif isinstance(mixed_precision, str):
        mixed_precision = mixed_precision.upper()
        if mixed_precision == 'FULL':
            pass
        elif mixed_precision == 'DEFAULT':
            param_dtype = get_torch_dtype(precision)
            buffer_dtype = get_torch_dtype(precision)
        elif mixed_precision == 'PURE':
            param_dtype = get_torch_dtype(precision)
            reduce_dtype = get_torch_dtype(precision)
            buffer_dtype = get_torch_dtype(precision)
        else:
            raise ValueError(f'Unable to interpret mixed_precision={mixed_precision}')
    else:
        raise ValueError(f'Unable to interpret mixed_precision={mixed_precision}')

    mixed_precision = MixedPrecision(
        param_dtype=param_dtype,
        reduce_dtype=reduce_dtype,
        buffer_dtype=buffer_dtype,
        keep_low_precision_grads=keep_low_precision_grads,
    )

    return mixed_precision, param_dtype, reduce_dtype, buffer_dtype


def get_cpu_offload(cpu_offload=False):
    """Helper fn for configuring cpu_offload."""
    cpu_offload = CPUOffload(offload_params=True) if cpu_offload else None
    if cpu_offload is not None:
        raise ValueError('FSDP CPU Offload not supported yet.')
    return cpu_offload


def get_process_group(pg, process_group_cache=None):
    """Helper function for configuring and/or retrieving process groups."""
    warnings.warn(f'Instantiating FSDP with custom process groups is an experimental feature.')

    # Return regular process_groups as is, no cacheing
    if pg is None or isinstance(pg, ProcessGroup):
        return pg

    world_size = dist.get_world_size()
    local_world_size = dist.get_local_world_size()

    # Handle special str process_group cases
    if pg == 'self':
        pg = 'set1'
        warnings.warn(f"Converting process_group='self' to process_group='{pg}'")
    elif pg == 'node':
        pg = f'set{local_world_size}'
        warnings.warn(f"Converting process_group='node' to process_group='{pg}'")
    elif pg == 'local_rank_across_nodes':
        pg = f'mod{local_world_size}'
        warnings.warn(f"Converting process_group='local_rank_across_nodes' to process_group='{pg}'")

    # Handle str and Union[List[int], Tuple[int]] process_group cases
    if isinstance(pg, str) and pg.startswith('set'):
        k = int(pg.strip('set'))
        world_size = dist.get_world_size()
        if world_size % k != 0:
            raise RuntimeError(f'{world_size} must be divisible by set size ({k})')
        start = dist.get_global_rank() // k * k
        ranks = tuple(range(start, start + k))
    elif isinstance(pg, str) and pg.startswith('mod'):
        k = int(pg.strip('mod'))
        world_size = dist.get_world_size()
        if world_size % k != 0:
            raise RuntimeError(f'{world_size} must be divisible by mod ({k})')
        ranks = tuple(range(dist.get_global_rank() % k, world_size, k))
    elif isinstance(pg, (list, tuple)):
        ranks = tuple(pg)
    else:
        raise ValueError(f'Unsure how to setup process_group={pg}')

    if process_group_cache is not None and ranks in process_group_cache:
        warnings.warn(
            f'On rank={dist.get_global_rank()} using cached progress group with {ranks=}. ' +\
            'If the intention was to use a new process group, a new process group can be instantiated and passed' +\
            "in as an arguement (`'process_group': newly_instantiated_process_group_obect,`)"
        )
        return process_group_cache[ranks]

    warnings.warn(
        f'Composer is instantiating custom process groups with {ranks=} (on rank={dist.get_global_rank()}). ' +\
        'This is an experimental feature.'
    )

    ranks_per_subgroup_list = list(set(dist.all_gather_object(ranks)))
    current_group, _subgroups = distributed.distributed_c10d.new_subgroups_by_enumeration(ranks_per_subgroup_list)

    if process_group_cache is not None:
        process_group_cache[ranks] = current_group
    return current_group


def _custom_recursive_wrap(module: nn.Module,
                           auto_wrap_policy: Callable,
                           wrapper_cls: Callable,
                           ignored_modules: Set[nn.Module],
                           ignored_params: Set[nn.Parameter],
                           process_group_cache: Dict[Tuple[int], Any],
                           only_wrap_children: bool = False,
                           **kwargs: Any) -> Tuple[nn.Module, int]:
    """Updates FSDPs _recursive_wrap to enable module_kwargs and custom process_group cache.

    modified version of
    https://github.com/pytorch/pytorch/blob/d922c29a22e4bf0fba49526f7536395eb8cd66f4/torch/distributed/fsdp/wrap.py#L353
    which recursively wraps modules as FSDP modules for parameter sharding.
    This modification enables the user to pass custom FSDP arguements for every wrapped module.
    The added process_group_cache enables different FSDP modules to, when appropriate, use the
    same process group instead of instantiating a new process group.

    Automatically wrap child modules of *module* that meet the given
    criteria with :func:`auto_wrap`. Does not rely on _ConfigAutoWrap.

    Args:
        module (nn.Module):
            module to recursively wrap
        auto_wrap_policy (Callable):
            A callable specifying a policy to recursively wrap layers with FSDP.
        ignored_modules (Set[torch.nn.Module]): Modules to ignore when
            wrapping.
        ignored_params (Set[torch.nn.Parameter]): Parameters to ignore when
            wrapping; these should be the parameters contained in the modules
            in ``ignored_modules``.
        process_group_cache (Dict[Tuple[int], Any]): a cache of process_group to
            use instead of potentially instantiating a new process_group

    Returns:
        (nn.Module, int):
            Wrapped module and the number parameters wrapped recursively.
    """
    assert auto_wrap_policy is not None, 'Must specify auto_wrap_policy.'
    assert wrapper_cls is not None, 'Must specify wrapper_cls'
    # Make sure no child is already wrapped.
    for _, child in module.named_modules():
        if child in ignored_modules:
            continue
        try:
            assert not isinstance(child, cast(type, wrapper_cls))
        except TypeError:
            # wrapper_cls is a function as opposed to a class type, just bypass above check.
            pass

    # We count all params, assuming none of them are already wrapped.
    num_params = sum(p.numel() for p in module.parameters() if p not in ignored_params)

    assert auto_wrap_policy is not None
    if auto_wrap_policy(module=module, recurse=True, unwrapped_params=num_params):
        total_wrapped_params = 0
        # Iterate through the children, recursively wrap if necessary
        for name, child in module.named_children():
            if child in ignored_modules:
                continue
            wrapped_child, num_wrapped_params = _custom_recursive_wrap(
                module=child,
                auto_wrap_policy=auto_wrap_policy,
                wrapper_cls=wrapper_cls,
                ignored_modules=ignored_modules,
                ignored_params=ignored_params,
                process_group_cache=process_group_cache,
                **kwargs,
            )
            setattr(module, name, wrapped_child)
            # Keep track of how many parameters have been wrapped
            total_wrapped_params += num_wrapped_params
        # decide if we need to wrap the current module,
        # since the left over parameters exceed the number of params to wrap
        remainder = num_params - total_wrapped_params
        module_kwargs = auto_wrap_policy(module=module, recurse=False, unwrapped_params=remainder)
        if not only_wrap_children and module_kwargs:
            module_kwargs = module_kwargs if isinstance(module_kwargs, dict) else {}
            # backward_prefetch_map
            if 'sharding_strategy' in module_kwargs and module_kwargs['sharding_strategy'] not in sharding_map.values():
                module_kwargs['sharding_strategy'] = sharding_map[module_kwargs['sharding_strategy'].upper()]
            if 'backward_prefetch' in module_kwargs and module_kwargs[
                    'backward_prefetch'] not in backward_prefetch_map.values():
                module_kwargs['backward_prefetch'] = backward_prefetch_map[module_kwargs['backward_prefetch'].upper()]
            if 'cpu_offload' in module_kwargs and not isinstance(module_kwargs['cpu_offload'], CPUOffload):
                module_kwargs['cpu_offload'] = get_cpu_offload(cpu_offload=module_kwargs['cpu_offload'].upper())
            if 'mixed_precision' in module_kwargs and not isinstance(module_kwargs['mixed_precision'], MixedPrecision):
                # `precision` needs to set `'mixed_precision'`, but `precision` is not part of fsdp kwargs
                raise NotImplementedError(
                    f"Automated setting of custom per module mixed_precision is not implemented, but it can be set if `isinstance(module_kwargs['mixed_precision'], MixedPrecision)`"
                )
            if 'process_group' in module_kwargs:
                module_kwargs['process_group'] = get_process_group(module_kwargs['process_group'], process_group_cache)

            final_kwargs = {**kwargs, **module_kwargs}

            # Leaf node or final wrapping of the remainder both happen here.
            return _wrap(module, wrapper_cls, **final_kwargs), num_params
        else:
            return module, total_wrapped_params
    return module, 0


class MosaicFullyShardedDataParallel(FullyShardedDataParallel):
    """Updates FSDP's _auto_wrap to enable module_kwargs."""

    def _auto_wrap(
        self,
        auto_wrap_kwargs: Dict[str, Any],
        fsdp_kwargs: Dict[str, Any],
    ) -> None:
        """Updates _auto_wrap to enable module_kwargs.

        modified version of
        https://github.com/pytorch/pytorch/blob/d922c29a22e4bf0fba49526f7536395eb8cd66f4/torch/distributed/fsdp/fully_sharded_data_parallel.py#L1252
        FSDP's _auto_wrap recursively wraps modules as FSDP modules for parameter sharding.
        This modification enables the user to pass custom FSDP arguements for every wrapped module.
        The added process_group_cache enables different FSDP modules to, when appropriate, use the
        same process group instead of instantiating a new process group.

        Recursively auto wraps the root module given by the key "module" in
        ``auto_wrap_kwargs`` with the arguments in ``auto_wrap_kwargs`` and
        ``fsdp_kwargs``.
        Precondition: ``auto_wrap_policy`` contains the arguments expected by
        ``_recursive_wrap()``, where ``auto_wrap_policy`` is not ``None``.
        ``fsdp_kwargs`` contains all FSDP arguments except ``module``.
        """
        auto_wrap_policy = auto_wrap_kwargs['auto_wrap_policy']
        root_module = auto_wrap_kwargs['module']
        assert auto_wrap_policy is not None
        # For auto wrapping, submodules should not already be wrapped with FSDP
        # since double wrapping is not supported
        for module_name, module in root_module.named_modules():
            if isinstance(module, FullyShardedDataParallel):
                raise ValueError(f'Expected {module_name} to NOT be FullyShardedDataParallel '
                                 'if using an `auto_wrap_policy`')
        mixed_precision = fsdp_kwargs['mixed_precision']
        if mixed_precision is not None and _contains_batchnorm(root_module):
            _override_batchnorm_mixed_precision(root_module)
            auto_wrap_policy = functools.partial(_or_policy, policies=[_wrap_batchnorm_individually, auto_wrap_policy])
            warnings.warn('Both mixed precision and an `auto_wrap_policy` were specified '
                          'for FSDP, where the wrapped module has batch norm submodules. '
                          'The batch norm submodules will be wrapped as separate FSDP '
                          'instances with mixed precision disabled since some batch norm '
                          'kernels do not support low precision.')
            auto_wrap_kwargs['auto_wrap_policy'] = auto_wrap_policy
        auto_wrap_kwargs['process_group_cache'] = {}
        _custom_recursive_wrap(**auto_wrap_kwargs, **fsdp_kwargs)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""BraTS (Brain Tumor Segmentation) dataset.

Please refer to the `Brain Tumor Segmentation (BraTS) challenge <https://www.med.upenn.edu/cbica/brats2021/>`_ for more
details about this dataset.
"""

import glob
import os
import random

import numpy as np
import torch
import torch.utils.data
import torchvision

from composer.utils import MissingConditionalImportError, dist

PATCH_SIZE = [1, 192, 160]

__all__ = ['PytTrain', 'PytVal']


def build_brats_dataloader(datadir: str,
                           global_batch_size: int,
                           oversampling: float = 0.33,
                           is_train: bool = True,
                           drop_last: bool = True,
                           shuffle: bool = True,
                           **dataloader_kwargs):
    """Builds a BRaTS dataloader

    Args:
        global_batch_size (int): Global batch size.
        **dataloader_kwargs (Dict[str, Any]): Additional settings for the dataloader (e.g. num_workers, etc.)
    """
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()
    x_train, y_train, x_val, y_val = get_data_split(datadir)
    dataset = PytTrain(x_train, y_train, oversampling) if is_train else PytVal(x_val, y_val)
    collate_fn = None if is_train else _my_collate
    sampler = dist.get_sampler(dataset, drop_last=drop_last, shuffle=shuffle)

    return torch.utils.data.DataLoader(dataset=dataset,
                                       batch_size=batch_size,
                                       sampler=sampler,
                                       drop_last=drop_last,
                                       collate_fn=collate_fn,
                                       **dataloader_kwargs)


def _my_collate(batch):
    """Custom collate function to handle images with different depths."""
    data = [item[0] for item in batch]
    target = [item[1] for item in batch]

    return [torch.Tensor(data), torch.Tensor(target)]


def _coin_flip(prob):
    return random.random() < prob


def _random_augmentation(probability, augmented, original):
    condition = _coin_flip(probability)
    neg_condition = condition ^ True
    return condition * augmented + neg_condition * original


class Crop(object):

    def __call__(self, data, oversampling):
        img, lbl = data['image'], data['label']

        def randrange(max_range):
            return 0 if max_range == 0 else random.randrange(max_range)

        def get_cords(cord, idx):
            return cord[idx], cord[idx] + PATCH_SIZE[idx]

        def _rand_crop(image, label):
            ranges = [s - p for s, p in zip(image.shape[1:], PATCH_SIZE)]

            cord = [randrange(x) for x in ranges]
            low_x, high_x = get_cords(cord, 0)
            low_y, high_y = get_cords(cord, 1)
            image = image[:, low_x:high_x, low_y:high_y]
            label = label[:, low_x:high_x, low_y:high_y]
            return image, label, [low_x, high_x, low_y, high_y]

        def rand_foreg_cropd(image, label):

            import scipy.ndimage
            cl = np.random.choice(np.unique(label[label > 0]))
            foreg_slices = scipy.ndimage.find_objects(scipy.ndimage.measurements.label(label == cl)[0])
            foreg_slices = [x for x in foreg_slices if x is not None]
            slice_volumes = [np.prod([s.stop - s.start for s in sl]) for sl in foreg_slices]
            slice_idx = np.argsort(slice_volumes)[-2:]
            foreg_slices = [foreg_slices[i] for i in slice_idx]
            if not foreg_slices:
                return _rand_crop(image, label)
            foreg_slice = foreg_slices[random.randrange(len(foreg_slices))]
            low_x, high_x = adjust(foreg_slice, PATCH_SIZE, label, 1)
            low_y, high_y = adjust(foreg_slice, PATCH_SIZE, label, 2)
            image = image[:, low_x:high_x, low_y:high_y]
            label = label[:, low_x:high_x, low_y:high_y]
            return image, label, [low_x, high_x, low_y, high_y]

        def adjust(foreg_slice, patch_size, label, idx):

            diff = patch_size[idx - 1] - (foreg_slice[idx].stop - foreg_slice[idx].start)
            sign = -1 if diff < 0 else 1
            diff = abs(diff)
            ladj = randrange(diff)
            hadj = diff - ladj
            low = max(0, foreg_slice[idx].start - sign * ladj)
            high = min(label.shape[idx], foreg_slice[idx].stop + sign * hadj)
            diff = patch_size[idx - 1] - (high - low)
            if diff > 0 and low == 0:
                high += diff
            elif diff > 0:
                low -= diff
            return low, high

        if random.random() < oversampling:
            img, lbl, _ = rand_foreg_cropd(img, lbl)
        else:
            img, lbl, _ = _rand_crop(img, lbl)

        return {'image': img, 'label': lbl}


class Noise(object):

    def __call__(self, data, oversampling):
        img, lbl = data['image'], data['label']
        std = np.random.uniform(0.0, oversampling)
        noise = np.random.normal(0, scale=std, size=img.shape)
        img_noised = img + noise
        img = _random_augmentation(0.15, img_noised, img)

        return {'image': img, 'label': lbl}


class Blur(object):

    def __call__(self, data):
        img, lbl = data['image'], data['label']

        transf = torchvision.transforms.GaussianBlur(kernel_size=3, sigma=(0.5, 1.5))
        img_blured = transf(torch.Tensor(img)).numpy()
        img = _random_augmentation(0.15, img_blured, img)

        return {'image': img, 'label': lbl}


class Brightness(object):

    def __call__(self, data):
        img, lbl = data['image'], data['label']
        brightness_scale = _random_augmentation(0.15, np.random.uniform(0.7, 1.3), 1.0)
        img = img * brightness_scale

        return {'image': img, 'label': lbl}


class Contrast(object):

    def __call__(self, data):
        img, lbl = data['image'], data['label']
        min_, max_ = np.min(img), np.max(img)
        scale = _random_augmentation(0.15, np.random.uniform(0.65, 1.5), 1.0)

        img = torch.clamp(torch.Tensor(img * scale), min_, max_).numpy()
        return {'image': img, 'label': lbl}


class Flips(object):

    def __call__(self, data):
        img, lbl = data['image'], data['label']
        axes = [1, 2]
        prob = 1 / len(axes)

        for axis in axes:
            if random.random() < prob:
                img = np.flip(img, axis=axis).copy()
                lbl = np.flip(lbl, axis=axis).copy()

        return {'image': img, 'label': lbl}


class Transpose(object):

    def __call__(self, data):
        img, lbl = data['image'], data['label']
        img, lbl = img.transpose((1, 0, 2, 3)), lbl.transpose((1, 0, 2, 3))

        return {'image': img, 'label': lbl}


class PytTrain(torch.utils.data.Dataset):

    def __init__(self, images, labels, oversampling, transform=None):
        self.images, self.labels = images, labels
        self.oversampling = oversampling
        self.transform = transform
        self.rand_crop = Crop()
        self.transpose = Transpose()
        self.contrast = Contrast()
        self.noise = Noise()
        self.blur = Blur()
        self.flips = Flips()
        self.bright = Brightness()

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        data = {'image': np.load(self.images[idx]), 'label': np.load(self.labels[idx])}
        data = self.rand_crop(data, self.oversampling)
        data = self.flips(data)
        data = self.noise(data, self.oversampling)
        data = self.blur(data)
        data = self.bright(data)
        data = self.contrast(data)
        data = self.transpose(data)

        return data['image'], data['label']


class PytVal(torch.utils.data.Dataset):

    def __init__(self, images, labels):
        self.images, self.labels = images, labels

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        data = {'image': np.load(self.images[idx]), 'label': np.load(self.labels[idx])}
        return data['image'], data['label']


def load_data(path, files_pattern):
    data = sorted(glob.glob(os.path.join(path, files_pattern)))
    assert len(data) > 0, f'Found no data at {path}'
    return data


def get_split(data, idx):
    return list(np.array(data)[idx])


def get_data_split(path: str):
    try:
        from sklearn.model_selection import KFold
    except ImportError as e:
        raise MissingConditionalImportError(extra_deps_group='unet',
                                            conda_channel='conda-forge',
                                            conda_package='scikit-learn') from e
    kfold = KFold(n_splits=5, shuffle=True, random_state=0)
    imgs = load_data(path, '*_x.npy')
    lbls = load_data(path, '*_y.npy')
    assert len(imgs) == len(lbls), f'Found {len(imgs)} volumes but {len(lbls)} corresponding masks'
    train_imgs, train_lbls, val_imgs, val_lbls = [], [], [], []

    train_idx, val_idx = list(kfold.split(imgs))[0]
    train_imgs = get_split(imgs, train_idx)
    train_lbls = get_split(lbls, train_idx)
    val_imgs = get_split(imgs, val_idx)
    val_lbls = get_split(lbls, val_idx)

    return train_imgs, train_lbls, val_imgs, val_lbls



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core Layer Freezing classes and functions."""

from __future__ import annotations

import logging
import textwrap
import warnings
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import torch
from torch.optim import Optimizer

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.utils import ensure_tuple

log = logging.getLogger(__name__)

__all__ = ['LayerFreezing', 'freeze_layers']


def freeze_layers(
    model: torch.nn.Module,
    optimizers: Union[Optimizer, Sequence[Optimizer]],
    current_duration: float,
    freeze_start: float = 0.5,
    freeze_level: float = 1.0,
) -> Tuple[int, float]:
    """Progressively freeze the layers of the network in-place
    during training, starting with the earlier layers.

    Example:
         .. testcode::

            from composer.algorithms.layer_freezing import freeze_layers
            freeze_depth, feeze_level = freeze_layers(
                                            model=model,
                                            optimizers=optimizer,
                                            current_duration=0.5,
                                            freeze_start=0.0,
                                            freeze_level=1.0
                                        )


    Args:
        model (torch.nn.Module): The model being trained.
        optimizers (torch.optim.Optimizer | Sequence[torch.optim.Optimizer]):
            The optimizers used during training.
        current_duration (float): The fraction, in ``[0, 1)`` of the training process complete.
        freeze_start (float, optional): The fraction of the training process in ``[0, 1)`` to run
            before freezing begins. Default: ``0.5``.
        freeze_level (float, optional): The maximum fraction of layers on ``[0, 1)`` to freeze.
            Default: ``1.0``.

    Return:
        (int, float): The number of layers frozen, and the percentage of the total model frozen.
    """
    # Flatten out the layers
    flat_children = []
    _get_layers(model, flat_children)
    # Determine how many layers to freeze
    freeze_percentage = _freeze_schedule(current_duration=current_duration,
                                         freeze_start=freeze_start,
                                         freeze_level=freeze_level)
    freeze_depth = int(freeze_percentage * len(flat_children[0:-1]))

    # Freeze the parameters in the chosen layers
    for i, child in enumerate(flat_children[0:-1]):
        if i < freeze_depth:
            for p in child.parameters():
                _remove_param_from_optimizers(p, optimizers)
                # Do not compute gradients for this param.
                p.requires_grad = False

    # Log results
    log.info(
        textwrap.dedent(f"""\
            Applied Layer Freezing with freeze_start={freeze_start},
            freeze_level={freeze_level}. Froze {freeze_depth} layers in the model which
            equates to {freeze_percentage * 100}% of all layers."""))
    return freeze_depth, freeze_percentage


class LayerFreezing(Algorithm):
    """Progressively freeze the layers of the network during training, starting with the earlier layers.

    Freezing starts after the fraction of training specified by ``freeze_start``
    has elapsed. The fraction of layers frozen increases linearly until it
    reaches ``freeze_level`` at the end of training.

    This freezing schedule is most similar to
    `FreezeOut <https://arxiv.org/abs/1706.04983>`_ and
    `Freeze Training <https://arxiv.org/abs/1706.05806>`_.

    Runs on :attr:`.Event.EPOCH_END`.

    Example:
         .. testcode::

            from composer.algorithms import LayerFreezing
            from composer.trainer import Trainer
            layer_freezing_algorithm = LayerFreezing(
                freeze_start=0.0,
                freeze_level=1.0
            )
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[layer_freezing_algorithm],
                optimizers=[optimizer]
            )

    Args:
        freeze_start (float): The fraction of training to run before freezing begins. Default: ``0.5``.
        freeze_level (float): The maximum fraction of layers to freeze. Default: ``1.0``.
    """

    def __init__(self, freeze_start: float = 0.5, freeze_level: float = 1.0):
        self.freeze_start = freeze_start
        self.freeze_level = freeze_level

    @property
    def find_unused_parameters(self) -> bool:
        """Override in order to tell DDP that some parameters will not have gradients computed for them after layer
        freezing is applied."""
        return True

    def match(self, event: Event, state: State) -> bool:
        del state  # unused
        return event == Event.EPOCH_END

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        del event  # unused
        optimizers = state.optimizers
        assert optimizers is not None
        elapsed_duration = state.get_elapsed_duration()
        assert elapsed_duration is not None, 'elapsed duration should be set on Event.EPOCH_END'
        freeze_depth, freeze_percentage = freeze_layers(
            model=state.model,
            optimizers=optimizers,
            current_duration=float(elapsed_duration),
            freeze_start=self.freeze_start,
            freeze_level=self.freeze_level,
        )
        logger.log_metrics({
            'layer_freezing/layers_frozen': freeze_depth,
            'layer_freezing/percentage_frozen': freeze_percentage
        })

    def state_dict(self) -> Dict[str, Any]:
        warnings.warn(('Checkpoints with layer freezing cannot reliably be used to resume training.'
                       'See: https://github.com/mosaicml/composer/issues/1002'))
        return {}

    def load_state_dict(self, state: Dict[str, Any]) -> None:
        warnings.warn(('Checkpoints with layer freezing cannot reliably be used to resume training.'
                       'See: https://github.com/mosaicml/composer/issues/1002'))


def _freeze_schedule(current_duration: float, freeze_start: float, freeze_level: float) -> float:
    """Implements a linear schedule for freezing.

    The schedule is linear and begins with no freezing and linearly
    increases the fraction of layers frozen, reaching the fraction specified by ``freeze_level`` at the end of training.
    The start of freezing is given as a fraction of the total training duration and is set with ``freeze_start``.

    Args:
        current_duration (float): The elapsed training duration.
        freeze_start (float): The fraction of training to run before freezing begins.
        freeze_level (float): The maximum fraction of levels to freeze.
    """
    # No freezing if the current epoch is less than this
    if current_duration <= freeze_start:
        return 0.0
    # `Calculate the total time for freezing to occur
    total_freezing_time = 1.0 - freeze_start
    # Calculate the amount of freezing time that has elapsed
    freezing_time_elapsed = current_duration - freeze_start
    # Calculate the fraction of the freezing time elapsed.
    freezing_time_elapsed_frac = freezing_time_elapsed / total_freezing_time
    # Scale this fraction by the amount of freezing to do.
    return freeze_level * freezing_time_elapsed_frac


def _get_layers(module: torch.nn.Module, flat_children: List[torch.nn.Module]):
    """Helper function to get all submodules.

    Does a depth first search to flatten out modules which
    contain parameters.

    Args:
        module (torch.nn.Module): Current module to search.
        flat_children (List[torch.nn.Module]): List containing modules.
    """
    # Check if given module has no children and parameters.
    if (len(list(module.children())) == 0 and len(list(module.parameters())) > 0):
        flat_children.append(module)
    else:
        # Otherwise, continue the search over its children.
        for child in module.children():
            _get_layers(child, flat_children)


def _remove_param_from_optimizers(p: torch.nn.Parameter, optimizers: Union[Optimizer, Sequence[Optimizer]]):
    """Helper function to freeze the training of a parameter.

    To freeze a parameter, it must be removed from the optimizer,
    otherwise momentum and weight decay may still be applied.

    Args:
        p (torch.nn.Parameter): The parameter being frozen.
        optimizers (torch.optim.Optimizer | Sequence[torch.optim.Optimizer]): The optimizers used during training.
    """
    # Search over params in the optimizers to find and remove the
    # given param. Necessary due to the way params are stored.
    for optimizer in ensure_tuple(optimizers):
        for group in optimizer.param_groups:
            group['params'] = list(filter(lambda x: id(x) != id(p), group['params']))



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""GPT-2 model based on `Hugging Face GPT-2 <https://huggingface.co/docs/transformers/master/en/model_doc/gpt2>`_.

Implemented as a wrapper using :class:`.ComposerTrainer`.
"""

from __future__ import annotations

from typing import Optional

from composer.metrics.nlp import LanguageCrossEntropy, LanguagePerplexity
from composer.models.huggingface import HuggingFaceModel
from composer.utils.import_helpers import MissingConditionalImportError

__all__ = ['create_gpt2']


def create_gpt2(use_pretrained: Optional[bool] = False,
                pretrained_model_name: Optional[str] = None,
                model_config: Optional[dict] = None,
                tokenizer_name: Optional[str] = None,
                gradient_checkpointing: Optional[bool] = False):
    """Implements :class:`~composer.models.huggingface.HuggingFaceModel` to wrap `Hugging Face GPT-2 \
    transformers <https://huggingface.co/docs/transformers/master/en/model_doc/gpt2#overview>`_. Logs training and
    validation perplexity.

    From `Language Models are Unsupervised Multitask Learners <https://d4mucfpksywv.cloudfront.net/better-language-models/language-models.pdf>`_ (Radford et al, 2018).

    Args:

        gradient_checkpointing (bool, optional): Use gradient checkpointing. Default: ``False``.
        use_pretrained (bool, optional): Whether to initialize the model with the pretrained weights. Default: ``False``.
        model_config (dict): A dictionary providing a HuggingFace model configuration.
        tokenizer_name (str, optional): Tokenizer name used to preprocess the dataset
        and validate the models inputs.

        .. code-block::

            {
              "_name_or_path": "gpt2",
              "activation_function": "gelu_new",
              "architectures": ["GPT2LMHeadModel"],
              "attn_pdrop": 0.1,
              "bos_token_id": 50256,
              "embd_pdrop": 0.1,
              "eos_token_id": 50256,
              "initializer_range": 0.02,
              "layer_norm_epsilon": 1e-05,
              "model_type": "gpt2",
              "n_ctx": 1024,
              "n_embd": 768,
              "n_head": 12,
              "n_inner": null,
              "n_layer": 12,
              "n_positions": 1024,
              "reorder_and_upcast_attn": false,
              "resid_pdrop": 0.1,
              "scale_attn_by_inverse_layer_idx": false,
              "scale_attn_weights": true,
              "summary_activation": null,
              "summary_first_dropout": 0.1,
              "summary_proj_to_labels": true,
              "summary_type": "cls_index",
              "summary_use_proj": true,
              "task_specific_params": {
              "text-generation": {
              "do_sample": true,
              "max_length": 50 }
              },
              "transformers_version": "4.16.0",
              "use_cache": true,
              "vocab_size": 50257
            }

   To create a GPT-2 model for language modeling pretraining:

    .. testcode::

        from composer.models import create_gpt2

        composer_model = create_gpt2()

    """
    try:
        import transformers
    except ImportError as e:
        raise MissingConditionalImportError(extra_deps_group='nlp', conda_package='transformers') from e

    if not model_config:
        model_config = {}

    if not pretrained_model_name:
        pretrained_model_name = 'gpt2'

    if use_pretrained:
        assert transformers.AutoModelForCausalLM.from_pretrained is not None, 'AutoModelForCausalLM has from_pretrained method'
        model = transformers.AutoModelForCausalLM.from_pretrained(pretrained_model_name_or_path=pretrained_model_name,
                                                                  **model_config)
    else:
        config = transformers.AutoConfig.from_pretrained(pretrained_model_name, **model_config)
        assert transformers.AutoModelForCausalLM.from_config is not None, 'AutoModelForCausalLM has from_config method'
        model = transformers.AutoModelForCausalLM.from_config(config)

    if gradient_checkpointing:
        model.gradient_checkpointing_enable()  # type: ignore

    # setup the tokenizer
    if tokenizer_name:
        tokenizer = transformers.AutoTokenizer.from_pretrained(tokenizer_name)
    else:
        tokenizer = None

    return HuggingFaceModel(model=model,
                            tokenizer=tokenizer,
                            metrics=[LanguageCrossEntropy(), LanguagePerplexity()],
                            use_logits=True)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Shrinks targets towards a uniform distribution to counteract label noise. Introduced in `Rethinking the Inception
Architecture for Computer Vision <https://arxiv.org/abs/1512.00567>`_.

See the :doc:`Method Card </method_cards/label_smoothing>` for more details.
"""

from composer.algorithms.label_smoothing.label_smoothing import LabelSmoothing as LabelSmoothing
from composer.algorithms.label_smoothing.label_smoothing import smooth_labels as smooth_labels

__all__ = ['LabelSmoothing', 'smooth_labels']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""ImageNet classification streaming dataset.

The most widely used dataset for Image Classification algorithms. Please refer to the `ImageNet 2012 Classification
Dataset <http://image-net.org/>`_ for more details.
"""

import os
from typing import Any, Dict, List, Optional

import numpy as np
import torch
from torch.utils.data import DataLoader
from torchvision import transforms
from torchvision.datasets import ImageFolder

from composer.core import DataSpec, MemoryFormat
from composer.datasets.ffcv_utils import ffcv_monkey_patches, write_ffcv_dataset
from composer.datasets.synthetic import SyntheticBatchPairDataset
from composer.datasets.utils import NormalizationFn, pil_image_collate
from composer.utils import MissingConditionalImportError, dist

__all__ = [
    'build_imagenet_dataloader',
    'build_streaming_imagenet1k_dataloader',
    'build_synthetic_imagenet_dataloader',
    'write_ffcv_imagenet',
    'build_ffcv_imagenet_dataloader',
]

IMAGENET_CHANNEL_MEAN = (0.485 * 255, 0.456 * 255, 0.406 * 255)
IMAGENET_CHANNEL_STD = (0.229 * 255, 0.224 * 255, 0.225 * 255)


def build_imagenet_dataloader(
    datadir: str,
    global_batch_size: int,
    is_train: bool = True,
    drop_last: bool = True,
    shuffle: bool = True,
    resize_size: int = -1,
    crop_size: int = 224,
    **dataloader_kwargs: Dict[str, Any],
) -> DataSpec:
    """Builds an ImageNet dataloader.

    Args:
        datadir (str): path to location of dataset.
        global_batch_size (int): Global batch size.
        is_train (bool): Whether to load the training data or validation data. Default:
            ``True``.
        drop_last (bool): whether to drop last samples. Default: ``True``.
        shuffle (bool): whether to shuffle the dataset. Default: ``True``.
        resize_size (int, optional): The resize size to use. Use ``-1`` to not resize. Default: ``-1``.
        crop size (int): The crop size to use. Default: ``224``.
        **dataloader_kwargs (Dict[str, Any]): Additional settings for the dataloader (e.g. num_workers, etc.)
    """
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()
    if is_train:
        # include fixed-size resize before RandomResizedCrop in training only
        # if requested (by specifying a size > 0)
        train_transforms: List[torch.nn.Module] = []

        if resize_size > 0:
            train_transforms.append(transforms.Resize(resize_size))

        train_transforms += [
            transforms.RandomResizedCrop(crop_size, scale=(0.08, 1.0), ratio=(0.75, 4.0 / 3.0)),
            transforms.RandomHorizontalFlip()
        ]
        transformation = transforms.Compose(train_transforms)
        split = 'train'
    else:
        val_transforms: List[torch.nn.Module] = []
        if resize_size > 0:
            val_transforms.append(transforms.Resize(resize_size))
        val_transforms.append(transforms.CenterCrop(crop_size))
        transformation = transforms.Compose(val_transforms)
        split = 'val'

    device_transform_fn = NormalizationFn(mean=IMAGENET_CHANNEL_MEAN, std=IMAGENET_CHANNEL_STD)

    dataset = ImageFolder(os.path.join(datadir, split), transformation)
    sampler = dist.get_sampler(dataset, drop_last=drop_last, shuffle=shuffle)

    return DataSpec(
        DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            sampler=sampler,
            drop_last=drop_last,
            collate_fn=pil_image_collate,
            **dataloader_kwargs,
        ),
        device_transforms=device_transform_fn,
    )


def build_synthetic_imagenet_dataloader(
    global_batch_size: int,
    num_unique_samples: int = 100,
    device: str = 'cpu',
    memory_format: MemoryFormat = MemoryFormat.CONTIGUOUS_FORMAT,
    is_train: bool = True,
    crop_size: int = 224,
    drop_last: bool = True,
    shuffle: bool = True,
    **dataloader_kwargs: Dict[str, Any],
) -> DataSpec:
    """Builds a synthetic ImageNet dataloader.

    Args:
        global_batch_size (int): Global batch size.
        num_unique_samples (int): number of unique samples in synthetic dataset. Default: ``100``.
        device (str): device with which to load the dataset. Default: ``cpu``.
        memory_format (:class:`composer.core.MemoryFormat`): memory format of the tensors. Default: ``CONTIGUOUS_FORMAT``.
        is_train (bool): Whether to load the training data or validation data. Default:
            ``True``.
        crop size (int): The crop size to use. Default: ``224``.
        drop_last (bool): whether to drop last samples. Default: ``True``.
        shuffle (bool): whether to shuffle the dataset. Default: ``True``.
        **dataloader_kwargs (Dict[str, Any]): Additional settings for the dataloader (e.g. num_workers, etc.)
    """
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()
    total_dataset_size = 1_281_167 if is_train else 50_000
    dataset = SyntheticBatchPairDataset(
        total_dataset_size=total_dataset_size,
        data_shape=[3, crop_size, crop_size],
        num_classes=1000,
        num_unique_samples_to_create=num_unique_samples,
        device=device,
        memory_format=memory_format,
    )

    sampler = dist.get_sampler(dataset, drop_last=drop_last, shuffle=shuffle)

    return DataSpec(
        DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            sampler=sampler,
            drop_last=drop_last,
            collate_fn=pil_image_collate,
            **dataloader_kwargs,
        ),)


def write_ffcv_imagenet(
    datadir: str,
    savedir: str = '/tmp/imagenet_train.ffcv',
    split: str = 'train',
    num_workers: int = 8,
):
    """Converts an ImageNet dataset to FFCV format.

        datadir (str): Path of ImageNet dataset, in ImageFolder format.
        savedir (str): Path to save the FFCV dataset. Default: ``/tmp/imagenet_train.ffcv``.
        split (str): 'train' or 'val'. Default: ``train``.
        num_workers (int): Number of workers to use for conversion. Default: ``8``.
    """

    if dist.get_local_rank() == 0:
        ds = ImageFolder(os.path.join(datadir, split))
        write_ffcv_dataset(dataset=ds,
                           write_path=savedir,
                           max_resolution=500,
                           num_workers=num_workers,
                           compress_probability=0.50,
                           jpeg_quality=90)

    # wait for rank 0 to finish conversion
    dist.barrier()


def build_ffcv_imagenet_dataloader(
    datadir: str,
    global_batch_size: int,
    is_train: bool = True,
    resize_size: int = -1,
    crop_size: int = 224,
    cpu_only: bool = False,
    drop_last: bool = True,
    prefetch_factor: int = 2,
    num_workers: int = 8,
):
    """Builds an FFCV ImageNet dataloader.

    Args:
        datadir (str): path to location of dataset.
        global_batch_size (int): Global batch size.
        is_train (bool): Whether to load the training data or validation data. Default:
            ``True``.
        resize_size (int, optional): The resize size to use. Use ``-1`` to not resize. Default: ``-1``.
        crop size (int): The crop size to use. Default: ``224``.
        cpu_only (int): Only perform transforms on 'cpu'. Default: ``False``.
        drop_last (bool): whether to drop last samples. Default: ``True``.
        prefetch_factor (int): Number of batches to prefect. Default: ``2``.
        num_workers (int): Number of workers. Default: ``8``.
    """
    try:
        import ffcv
        from ffcv.fields.decoders import CenterCropRGBImageDecoder, IntDecoder, RandomResizedCropRGBImageDecoder
        from ffcv.pipeline.operation import Operation
    except ImportError:
        raise ImportError('Composer was installed without ffcv support.'
                          'To use ffcv with Composer, please install ffcv.')
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()
    device = torch.device(f'cuda:{dist.get_local_rank()}')
    label_pipeline: List[Operation] = [
        IntDecoder(),
        ffcv.transforms.ToTensor(),
        ffcv.transforms.Squeeze(),
        ffcv.transforms.ToDevice(device, non_blocking=True)
    ]
    image_pipeline: List[Operation] = []
    if is_train:
        image_pipeline.extend(
            [RandomResizedCropRGBImageDecoder((crop_size, crop_size)),
             ffcv.transforms.RandomHorizontalFlip()])
        dtype = np.float16
    else:
        ratio = crop_size / resize_size if resize_size > 0 else 1.0
        image_pipeline.extend([CenterCropRGBImageDecoder((crop_size, crop_size), ratio=ratio)])
        dtype = np.float32

    # Common transforms for train and test
    if cpu_only:
        image_pipeline.extend([
            ffcv.transforms.NormalizeImage(np.array(IMAGENET_CHANNEL_MEAN), np.array(IMAGENET_CHANNEL_STD), dtype),
            ffcv.transforms.ToTensor(),
            ffcv.transforms.ToTorchImage(),
        ])
    else:
        image_pipeline.extend([
            ffcv.transforms.ToTensor(),
            ffcv.transforms.ToDevice(device, non_blocking=True),
            ffcv.transforms.ToTorchImage(),
            ffcv.transforms.NormalizeImage(np.array(IMAGENET_CHANNEL_MEAN), np.array(IMAGENET_CHANNEL_STD), dtype),
        ])

    is_distributed = dist.get_world_size() > 1

    ffcv_monkey_patches()
    ordering = ffcv.loader.OrderOption.RANDOM if is_train else ffcv.loader.OrderOption.SEQUENTIAL

    return ffcv.Loader(
        datadir,
        batch_size=batch_size,
        num_workers=num_workers,
        order=ordering,
        distributed=is_distributed,
        pipelines={
            'image': image_pipeline,
            'label': label_pipeline
        },
        batches_ahead=prefetch_factor,
        drop_last=drop_last,
    )


def build_streaming_imagenet1k_dataloader(
    global_batch_size: int,
    remote: str,
    *,
    local: str = '/tmp/mds-cache/mds-imagenet1k',
    split: str = 'train',
    drop_last: bool = True,
    shuffle: bool = True,
    resize_size: int = -1,
    crop_size: int = 224,
    predownload: Optional[int] = 100_000,
    keep_zip: Optional[bool] = None,
    download_retry: int = 2,
    download_timeout: float = 60,
    validate_hash: Optional[str] = None,
    shuffle_seed: Optional[int] = None,
    num_canonical_nodes: Optional[int] = None,
    **dataloader_kwargs: Dict[str, Any],
) -> DataSpec:
    """Builds an imagenet1k streaming dataset

    Args:
        global_batch_size (int): Global batch size.
        remote (str): Remote directory (S3 or local filesystem) where dataset is stored.
        local (str, optional): Local filesystem directory where dataset is cached during operation.
            Defaults to ``'/tmp/mds-cache/mds-imagenet1k/```.
        split (str): Which split of the dataset to use. Either ['train', 'val']. Default:
            ``'train```.
        drop_last (bool, optional): whether to drop last samples. Default: ``True``.
        shuffle (bool, optional): whether to shuffle dataset. Defaults to ``True``.
        resize_size (int, optional): The resize size to use. Use ``-1`` to not resize. Default: ``-1``.
        crop size (int): The crop size to use. Default: ``224``.
        predownload (int, optional): Target number of samples ahead to download the shards of while
            iterating. Defaults to ``100_000``.
        keep_zip (bool, optional): Whether to keep or delete the compressed file when
            decompressing downloaded shards. If set to None, keep iff remote is local. Defaults to
            ``None``.
        download_retry (int): Number of download re-attempts before giving up. Defaults to ``2``.
        download_timeout (float): Number of seconds to wait for a shard to download before raising
            an exception. Defaults to ``60``.
        validate_hash (str, optional): Optional hash or checksum algorithm to use to validate
            shards. Defaults to ``None``.
        shuffle_seed (int, optional): Seed for shuffling, or ``None`` for random seed. Defaults to
            ``None``.
        num_canonical_nodes (int, optional): Canonical number of nodes for shuffling with resumption.
            Defaults to ``None``, which is interpreted as the number of nodes of the initial run.
        **dataloader_kwargs (Dict[str, Any]): Additional settings for the dataloader (e.g. num_workers, etc.)
    """
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()

    try:
        from streaming.vision import StreamingImageNet
    except ImportError as e:
        raise MissingConditionalImportError(extra_deps_group='streaming', conda_package='mosaicml-streaming') from e

    transform = []
    if split == 'train':
        # include fixed-size resize before RandomResizedCrop in training only
        # if requested (by specifying a size > 0)
        if resize_size > 0:
            transform.append(transforms.Resize(resize_size))
        # always include RandomResizedCrop and RandomHorizontalFlip
        transform += [
            transforms.RandomResizedCrop(crop_size, scale=(0.08, 1.0), ratio=(0.75, 4.0 / 3.0)),
            transforms.RandomHorizontalFlip()
        ]
    else:
        if resize_size > 0:
            transform.append(transforms.Resize(resize_size))
        transform.append(transforms.CenterCrop(crop_size))
    transform.append(lambda image: image.convert('RGB'))
    transform = transforms.Compose(transform)

    dataset = StreamingImageNet(
        local=local,
        remote=remote,
        split=split,
        shuffle=shuffle,
        transform=transform,
        predownload=predownload,
        keep_zip=keep_zip,
        download_retry=download_retry,
        download_timeout=download_timeout,
        validate_hash=validate_hash,
        shuffle_seed=shuffle_seed,
        num_canonical_nodes=num_canonical_nodes,
        batch_size=batch_size,
    )

    dataloader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        collate_fn=pil_image_collate,
        drop_last=drop_last,
        **dataloader_kwargs,
    )

    device_transform_fn = NormalizationFn(mean=IMAGENET_CHANNEL_MEAN, std=IMAGENET_CHANNEL_STD)
    return DataSpec(dataloader=dataloader, device_transforms=device_transform_fn)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The models module contains the :class:`.ComposerModel` base class along with reference
implementations of many common models. Additionally, it includes task-specific convenience
:class:`.ComposerModel`\\s that wrap existing Pytorch models with standard forward passes
and logging to enable quick interaction with the :class:`.Trainer`.

See :doc:`Composer Model </composer_model>` for more details.
"""

from composer.models.base import ComposerModel
from composer.models.bert import create_bert_classification, create_bert_mlm
from composer.models.classify_mnist import mnist_model
from composer.models.deeplabv3 import composer_deeplabv3
from composer.models.efficientnetb0 import composer_efficientnetb0
from composer.models.gpt2 import create_gpt2
from composer.models.huggingface import HuggingFaceModel, write_huggingface_pretrained_from_composer_checkpoint
from composer.models.initializers import Initializer
from composer.models.mmdetection import MMDetModel
from composer.models.resnet import composer_resnet
from composer.models.resnet_cifar import composer_resnet_cifar
from composer.models.tasks import ComposerClassifier
from composer.models.timm import composer_timm
from composer.models.unet import UNet
from composer.models.vit_small_patch16 import vit_small_patch16

__all__ = [
    'ComposerModel',
    'create_bert_classification',
    'create_bert_mlm',
    'mnist_model',
    'composer_deeplabv3',
    'composer_efficientnetb0',
    'create_gpt2',
    'HuggingFaceModel',
    'write_huggingface_pretrained_from_composer_checkpoint',
    'Initializer',
    'MMDetModel',
    'composer_resnet',
    'composer_resnet_cifar',
    'ComposerClassifier',
    'composer_timm',
    'UNet',
    'vit_small_patch16',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import os
import platform

import PIL
import pytest
from torch.utils.data import DataLoader

from composer.trainer import Trainer
from tests.common import RandomClassificationDataset, SimpleModel


@pytest.mark.skipif('composer-python' not in os.environ['PATH'] or 'Linux' not in platform.system(),
                    reason='Pillow-simd test only checks if using the composer docker')
class TestDocker:

    def test_pillow_simd(self):
        assert 'post' in PIL.__version__, 'pillow-simd is not installed'

    @pytest.mark.gpu
    def test_apex(self):
        """Test that apex is installed and works in the GPU image."""
        import apex

        model = SimpleModel()
        opt = apex.optimizers.FusedAdam(model.parameters(), lr=0.01)
        trainer = Trainer(
            model=model,
            train_dataloader=DataLoader(RandomClassificationDataset()),
            optimizers=opt,
            max_duration='2ba',
        )

        trainer.fit()



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Create a submission for MLPerf Training benchmark."""

import json
import logging
import os
import platform
import subprocess
import sys
import warnings
from typing import Any, Dict, Iterable, Optional

import torch
from torch.utils.data import DataLoader, IterableDataset

import composer
from composer.core import Callback, State
from composer.loggers import Logger
from composer.utils import dist

try:
    import cpuinfo
    import psutil
    from mlperf_logging import mllog
    from mlperf_logging.mllog import constants

    mlperf_available = True
except ImportError:
    mlperf_available = False

# this callback only supports the following options:
BENCHMARKS = ('resnet', 'bert')
DIVISIONS = ('open',)
STATUS = ('onprem', 'cloud', 'preview')

__all__ = ['MLPerfCallback', 'get_system_description']


def _global_rank_zero() -> bool:
    return dist.get_global_rank() == 0


def _local_rank_zero() -> bool:
    return dist.get_local_rank() == 0


def _require_mlperf_logging():
    if not mlperf_available:
        raise ImportError("""Please install with `pip install 'mosaicml[mlperf]'` and also
                          install the logging library from: https://github.com/mlcommons/logging""")


class MLPerfCallback(Callback):
    """Create compliant results file for MLPerf Training benchmark.

    A submission folder structure will be created with the ``root_folder``
    as the base and the following directories::

        root_folder/
            results/
                [system_name]/
                    [benchmark]/
                        results_0.txt
                        results_1.txt
                        ...
            systems/
                [system_name].json

    A required systems description will be automatically generated,
    and best effort made to populate the fields, but should be manually
    checked prior to submission.

    Currently, only open division submissions are supported with this Callback.

    Example:
        .. code-block:: python

            from composer.callbacks import MLPerfCallback

            callback = MLPerfCallback(
                root_folder='/submission',
                index=0,
                metric_name='MulticlassAccuracy',
                metric_label='eval',
                target='0.759',
            )

    During training, the metric found in ``state.eval_metrics[metric_label][metric_name]``
    will be compared against the target criterion.

    .. note::

        This is currently an experimental logger that has not been used (yet)
        to submit an actual result to MLPerf. Please use with caution.

    .. note::

        MLPerf submissions require clearing the system cache prior to any training run.
        By default, this callback does not clear the cache, as that is a system specific
        operation. To enable cache clearing, and thus pass the mlperf compliance checker,
        provide a ``cache_clear_cmd`` that will be executed with ``os.system``.

    Args:
        root_folder (str): The root submission folder
        index (int): The repetition index of this run. The filename created will be
            ``result_[index].txt``.
        benchmark (str, optional): Benchmark name. Currently only ``resnet`` supported.
            Default: ``'resnet'``.
        target (float, optional): The target metric before the mllogger marks the stop
            of the timing run. Default: ``0.759`` (resnet benchmark).
        division (str, optional): Division of submission. Currently only ``open`` division supported.
            Default: ``'open'``.
        metric_name (str, optional): name of the metric to compare against the target.
            Default: ``MulticlassAccuracy``.
        metric_label (str, optional): The label name. The metric will be accessed via
            ``state.eval_metrics[metric_label][metric_name]``.
        submitter (str, optional): Submitting organization. Default: ``"MosaicML"``.
        system_name (str, optional): Name of the system (e.g. 8xA100_composer). If
            not provided, system name will default to ``[world_size]x[device_name]_composer``,
            e.g. ``8xNVIDIA_A100_80GB_composer``.
        status (str, optional): Submission status. One of (onprem, cloud, or preview).
            Default: ``"onprem"``.
        cache_clear_cmd (str, optional): Command to invoke during the cache clear. This callback
            will call ``os.system(cache_clear_cmd)``. Default is disabled (None)
        host_processors_per_node (int, optional): Total number of host processors per node.  Default: ``None``.
        exit_at_target (bool, optional): Whether to exit training when target metric is met. Default: ``False``.
    """

    def __init__(
        self,
        root_folder: str,
        index: int,
        benchmark: str = 'resnet',
        target: float = 0.759,
        division: str = 'open',
        metric_name: str = 'MulticlassAccuracy',
        metric_label: str = 'eval',
        submitter: str = 'MosaicML',
        system_name: Optional[str] = None,
        status: str = 'onprem',
        cache_clear_cmd: Optional[str] = None,
        host_processors_per_node: Optional[int] = None,
        exit_at_target: bool = False,
    ) -> None:

        _require_mlperf_logging()

        if benchmark not in BENCHMARKS:
            raise ValueError(f'benchmark: {benchmark} must be one of {BENCHMARKS}')
        if division not in DIVISIONS:
            raise ValueError(f'division: {division} must be one of {DIVISIONS}')
        if status not in STATUS:
            raise ValueError(f'status: {status} must be one of {STATUS}')

        self.mllogger = mllog.get_mllogger()
        self.target = target
        self.benchmark = benchmark
        self.target = target
        self.division = division
        self.submitter = submitter
        self.status = status
        self.cache_clear_cmd = cache_clear_cmd
        self.root_folder = root_folder
        self.metric_name = metric_name
        self.metric_label = metric_label
        self.exit_at_target = exit_at_target
        self._file_handler = None

        self.system_desc = get_system_description(submitter, division, status, system_name, host_processors_per_node)
        if system_name is None:
            system_name = self.system_desc['system_name']
        self.system_name = system_name

        # file paths to save the systems file, results file
        self.systems_path = os.path.join(root_folder, 'systems', f'{system_name}.json')
        self.filename = os.path.join(root_folder, 'results', system_name, benchmark, f'result_{index}.txt')

        # upload names for object store logging
        self.upload_name = '{run_name}' + f'/results/{system_name}/{benchmark}/result_{index}.txt'
        self.system_desc_upload_name = '{run_name}' + f'/systems/{system_name}.json'

        self.success = False

    def init(self, state: State, logger: Logger) -> None:
        # setup here requies access to rank, which is only available after
        # the trainer is initialized
        if _local_rank_zero():
            self._create_submission_folders(self.root_folder, self.system_name, self.benchmark)
            with open(self.systems_path, 'w') as f:
                json.dump(self.system_desc, f, indent=4)

            if os.path.exists(self.filename):
                raise FileExistsError(f'{self.filename} already exists.')

        dist.barrier()

        self._file_handler = logging.FileHandler(self.filename)
        self._file_handler.setLevel(logging.INFO)
        self.mllogger.logger.addHandler(self._file_handler)

        if self.cache_clear_cmd is not None:
            if _local_rank_zero():
                subprocess.run(self.cache_clear_cmd.split(), check=True, text=True)
                self.mllogger.start(key=mllog.constants.CACHE_CLEAR)
        else:
            warnings.warn('cache_clear_cmd was not provided. For a valid submission, please provide the command.')

        dist.barrier()

        if _local_rank_zero():
            self.mllogger.start(key=mllog.constants.INIT_START)

        if _global_rank_zero():
            self._log_dict({
                constants.SUBMISSION_BENCHMARK: self.benchmark,
                constants.SUBMISSION_DIVISION: self.division,
                constants.SUBMISSION_ORG: self.submitter,
                constants.SUBMISSION_PLATFORM: self.system_name,
                constants.SUBMISSION_STATUS: self.status,
            })

            # optionally, upload the system description file
            logger.upload_file(self.system_desc_upload_name, self.systems_path)

    def _create_submission_folders(self, root_folder: str, system_name: str, benchmark: str):
        os.makedirs(root_folder, exist_ok=True)

        results_folder = os.path.join(root_folder, 'results')
        log_folder = os.path.join(root_folder, 'results', system_name)
        benchmark_folder = os.path.join(log_folder, benchmark)
        systems_folder = os.path.join(root_folder, 'systems')

        os.makedirs(results_folder, exist_ok=True)
        os.makedirs(log_folder, exist_ok=True)
        os.makedirs(benchmark_folder, exist_ok=True)
        os.makedirs(systems_folder, exist_ok=True)

    def _log_dict(self, data: Dict[str, Any]):
        for key, value in data.items():
            self.mllogger.event(key=key, value=value)

    def _get_accuracy(self, state: State) -> float:
        if self.metric_name not in state.eval_metrics[self.metric_label]:
            raise ValueError(f'{self.metric_name} must be a validation metric.')

        metric = state.eval_metrics[self.metric_label][self.metric_name].compute()
        return float(metric)

    def _get_time(self, state: State) -> int:
        """Different benchmarks log different units of time."""
        benchmark_time = {
            'resnet': state.timestamp.epoch.value,
            'bert': state.timestamp.sample.value,
        }
        return benchmark_time[self.benchmark]

    def _get_dataloader_stats(self, dataloader: Iterable):
        """Returns a tuple of ``(batch_size, num_samples)``."""
        if isinstance(dataloader, DataLoader):
            num_samples = len(dataloader.dataset)  # type: ignore
            if isinstance(dataloader.dataset, IterableDataset):
                num_samples *= dist.get_world_size()
            return (dataloader.batch_size, num_samples)
        try:
            # attempt to import ffcv and test if its an ffcv loader.
            import ffcv  # type: ignore

            if isinstance(dataloader, ffcv.loader.Loader):
                # Use the cached attribute ffcv.init_traversal_order to compute number of samples
                return (
                    dataloader.batch_size,  # type: ignore
                    len(dataloader.next_traversal_order()) * dist.get_world_size()  # type: ignore
                )
        except ImportError:
            pass

        raise TypeError(f'torch dataloader or ffcv dataloader required (and ffcv installed)')

    def fit_start(self, state: State, logger: Logger) -> None:
        if _global_rank_zero():
            if len(state.evaluators) > 1:
                raise ValueError('Only one evaluator is supported for the MLPerfCallback.')

            if state.train_dataloader is None:
                raise ValueError('Train dataloader need to be provided')

            batch_size, num_samples = self._get_dataloader_stats(state.train_dataloader)
            _, eval_num_samples = self._get_dataloader_stats(state.evaluators[0].dataloader.dataloader)

            if batch_size is None:
                raise ValueError('Batch size is required to be set for dataloader.')

            self._log_dict({
                constants.SEED: state.seed,
                constants.GLOBAL_BATCH_SIZE: batch_size * dist.get_world_size(),
                constants.GRADIENT_ACCUMULATION_STEPS: state.grad_accum,
                constants.TRAIN_SAMPLES: num_samples,
                constants.EVAL_SAMPLES: eval_num_samples,
            })

        if _local_rank_zero():
            self.mllogger.event(key=constants.INIT_STOP)

        dist.barrier()

        if _global_rank_zero():
            self.mllogger.event(key=constants.RUN_START)

    def epoch_start(self, state: State, logger: Logger) -> None:
        if _global_rank_zero():
            self.mllogger.event(key=constants.EPOCH_START, metadata={'epoch_num': self._get_time(state)})
            self.mllogger.event(key=constants.BLOCK_START,
                                metadata={
                                    'first_epoch_num': self._get_time(state),
                                    'epoch_count': 1
                                })

    def epoch_end(self, state: State, logger: Logger) -> None:
        if _global_rank_zero():
            self.mllogger.event(key=constants.EPOCH_STOP, metadata={'epoch_num': self._get_time(state)})
            logger.upload_file(remote_file_name=self.upload_name, file_path=self.filename)

    def eval_start(self, state: State, logger: Logger) -> None:
        if _global_rank_zero():
            self.mllogger.event(key=constants.EVAL_START, metadata={'epoch_num': self._get_time(state)})

    def eval_end(self, state: State, logger: Logger) -> None:
        accuracy = self._get_accuracy(state)

        if _global_rank_zero():
            self.mllogger.event(key=constants.EVAL_STOP, metadata={'epoch_num': self._get_time(state)})
            self.mllogger.event(key=constants.EVAL_ACCURACY,
                                value=accuracy,
                                metadata={'epoch_num': self._get_time(state)})
            self.mllogger.event(key=constants.BLOCK_STOP, metadata={'first_epoch_num': self._get_time(state)})

            if accuracy > self.target and not self.success:
                self.mllogger.event(key=constants.RUN_STOP, metadata={'status': 'success'})
                self.mllogger.logger.removeHandler(self._file_handler)
                self.success = True  # only log once

            # upload to object store after eval complete
            logger.upload_file(remote_file_name=self.upload_name, file_path=self.filename)

        if accuracy > self.target and self.exit_at_target:
            # stop training
            state.stop_training()

    def close(self, state: State, logger: Logger) -> None:
        if self._file_handler is not None:
            self._file_handler.close()


def get_system_description(
    submitter: str,
    division: str,
    status: str,
    system_name: Optional[str] = None,
    host_processors_per_node: Optional[int] = None,
) -> Dict[str, str]:
    """Generates a valid system description.

    Makes a best effort to auto-populate some of the fields, but should
    be manually checked prior to submission. The system name is
    auto-generated as ``"[world_size]x[device_name]_composer"``, e.g.
    ``"8xNVIDIA_A100_80GB_composer"``.

    Args:
        submitter (str): Name of the submitting organization.
        division (str): Submission division (open, closed).
        status (str): System status (cloud, onprem, preview).
        system_name (str, optional): System name. Default: ``None``.

    Returns:
        system description as a dictionary
    """
    is_cuda = torch.cuda.is_available()
    cpu_info = cpuinfo.get_cpu_info()

    system_desc = {
        'submitter': submitter,
        'division': division,
        'status': status,
        'number_of_nodes': dist.get_world_size() / dist.get_local_world_size(),
        'host_processors_per_node': str(host_processors_per_node) if host_processors_per_node else '',
        'host_processor_model_name': str(cpu_info.get('brand_raw', 'CPU')),
        'host_processor_core_count': str(psutil.cpu_count(logical=False)),
        'host_processor_vcpu_count': '',
        'host_processor_frequency': '',
        'host_processor_caches': '',
        'host_processor_interconnect': '',
        'host_memory_capacity': '',
        'host_storage_type': '',
        'host_storage_capacity': '',
        'host_networking': '',
        'host_networking_topology': '',
        'host_memory_configuration': '',
        'accelerators_per_node': str(dist.get_local_world_size()) if is_cuda else '0',
        'accelerator_model_name': str(torch.cuda.get_device_name(None)) if is_cuda else '',
        'accelerator_host_interconnect': '',
        'accelerator_frequency': '',
        'accelerator_on-chip_memories': '',
        'accelerator_memory_configuration': '',
        'accelerator_memory_capacity': '',
        'accelerator_interconnect': '',
        'accelerator_interconnect_topology': '',
        'cooling': '',
        'hw_notes': '',
        'framework':
            f'PyTorch v{torch.__version__} and MosaicML composer v{composer.__version__}',  # type: ignore (third-party missing stub)
        'other_software_stack': {
            'cuda_version': torch.version.cuda if is_cuda else '',  # type: ignore (third-party missing stub)
            'composer_version': composer.__version__,
            'python_version': sys.version,
        },
        'operating_system': f'{platform.system()} {platform.release()}',
        'sw_notes': '',
    }

    if system_desc['number_of_nodes'] != 1:
        warnings.warn('Number of nodes > 1 not tested, proceed with caution.')

    if system_name is None:
        world_size = dist.get_world_size()
        if is_cuda:
            device_name = system_desc['accelerator_model_name']
        else:
            device_name = system_desc['host_processor_model_name']

        device_name = device_name.replace(' ', '_')
        system_name = f'{world_size}x{device_name}_composer'

    # default to system name as "[world_size]x[device_name]"
    # e.g. 8xNVIDIA_A100_80GB
    system_desc['system_name'] = system_name
    return system_desc



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Optimizers and learning rate schedulers.

Composer is compatible with optimizers based off of PyTorch's native :class:`~torch.optim.Optimizer` API, and common
optimizers such
However, where applicable, it is recommended to use the optimizers provided in :mod:`~.decoupled_weight_decay` since
they improve off of their PyTorch equivalents.

PyTorch schedulers can be used with Composer, but this is explicitly discouraged. Instead, it is recommended to use
schedulers based off of Composer's :class:`~.scheduler.ComposerScheduler` API, which allows more flexibility and
configuration in writing schedulers.
"""

from composer.optim.decoupled_weight_decay import DecoupledAdamW, DecoupledSGDW
from composer.optim.scheduler import (ComposerScheduler, ConstantScheduler, ConstantWithWarmupScheduler,
                                      CosineAnnealingScheduler, CosineAnnealingWarmRestartsScheduler,
                                      CosineAnnealingWithWarmupScheduler, ExponentialScheduler, LinearScheduler,
                                      LinearWithWarmupScheduler, MultiStepScheduler, MultiStepWithWarmupScheduler,
                                      PolynomialScheduler, PolynomialWithWarmupScheduler, StepScheduler,
                                      compile_composer_scheduler)

__all__ = [
    'DecoupledAdamW',
    'DecoupledSGDW',
    'ComposerScheduler',
    'ConstantScheduler',
    'ConstantWithWarmupScheduler',
    'CosineAnnealingScheduler',
    'CosineAnnealingWarmRestartsScheduler',
    'CosineAnnealingWithWarmupScheduler',
    'ExponentialScheduler',
    'LinearScheduler',
    'LinearWithWarmupScheduler',
    'MultiStepScheduler',
    'MultiStepWithWarmupScheduler',
    'PolynomialScheduler',
    'PolynomialWithWarmupScheduler',
    'StepScheduler',
    'compile_composer_scheduler',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Base module for callbacks."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any

from composer.core.serializable import Serializable

if TYPE_CHECKING:
    from composer import Event, State
    from composer.loggers import Logger

__all__ = ['Callback']


class Callback(Serializable, abc.ABC):
    """Base class for callbacks.

    Callbacks provide hooks that can run at each training loop :class:`.Event`. A callback is similar to
    an :class:`.Algorithm` in that they are run on specific events, but it differs from an :class:`.Algorithm`
    in that it should not modify the training of the model. By convention, callbacks should not modify the
    :class:`.State`. They are typically used to for non-essential recording functions such as logging or timing.

    Callbacks can be implemented in two ways:

    #.  Override the individual methods named for each :class:`.Event`.

        For example,

        .. doctest::

            >>> class MyCallback(Callback):
            ...     def epoch_start(self, state: State, logger: Logger):
            ...         print(f'Epoch: {int(state.timestamp.epoch)}')
            >>> # construct trainer object with your callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[MyCallback()],
            ... )
            >>> # trainer will run MyCallback whenever the EPOCH_START
            >>> # is triggered, like this:
            >>> _ = trainer.engine.run_event(Event.EPOCH_START)
            Epoch: 0

    #.  Override :meth:`run_event` if you want a single method to handle all events. If this method is overridden, then
        the individual methods corresponding to each event name (such as :meth:`epoch_start`) will no longer be
        automatically invoked. For example, if you override :meth:`run_event`, then :meth:`epoch_start` will not be called
        on the :attr:`.Event.EPOCH_START` event, :meth:`batch_start` will not be called on the
        :attr:`.Event.BATCH_START`, etc. However, you can invoke :meth:`epoch_start`, :meth:`batch_start`, etc. in your
        overriding implementation of :meth:`run_event`.

        For example,

        .. doctest::

            >>> class MyCallback(Callback):
            ...     def run_event(self, event: Event, state: State, logger: Logger):
            ...         if event == Event.EPOCH_START:
            ...             print(f'Epoch: {int(state.timestamp.epoch)}')
            >>> # construct trainer object with your callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[MyCallback()],
            ... )
            >>> # trainer will run MyCallback whenever the EPOCH_START
            >>> # is triggered, like this:
            >>> _ = trainer.engine.run_event(Event.EPOCH_START)
            Epoch: 0
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # Stub signature for pyright
        del args, kwargs  # unused
        pass

    def run_event(self, event: Event, state: State, logger: Logger) -> None:
        """Called by the engine on each event.

        Args:
            event (Event): The event.
            state (State): The state.
            logger (Logger): The logger.
        """
        event_cb = getattr(self, event.value)
        return event_cb(state, logger)

    def init(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.INIT` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def after_load(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.AFTER_LOAD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def fit_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.FIT_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def epoch_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EPOCH_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def before_dataloader(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BEFORE_DATALOADER` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def after_dataloader(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.AFTER_DATALOADER` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def batch_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BATCH_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def before_train_batch(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BEFORE_TRAIN_BATCH` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def before_forward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BEFORE_FORWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def after_forward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.AFTER_FORWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def before_loss(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BEFORE_LOSS` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def after_loss(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.AFTER_LOSS` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def before_backward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BEFORE_BACKWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def after_backward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.AFTER_BACKWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def after_train_batch(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.AFTER_TRAIN_BATCH` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def batch_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BATCH_END` event.

        .. note::

           The following :attr:`.State.timestamp` member variables are
           incremented immediately before the :attr:`.Event.BATCH_END` event.

           +------------------------------------+
           | :attr:`.Timestamp.batch`           |
           +------------------------------------+
           | :attr:`.Timestamp.batch_in_epoch`  |
           +------------------------------------+
           | :attr:`.Timestamp.sample`          |
           +------------------------------------+
           | :attr:`.Timestamp.sample_in_epoch` |
           +------------------------------------+
           | :attr:`.Timestamp.token`           |
           +------------------------------------+
           | :attr:`.Timestamp.token_in_epoch`  |
           +------------------------------------+

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def batch_checkpoint(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.BATCH_CHECKPOINT` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def epoch_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EPOCH_END` event.

        .. note::

            :attr:`.State.timestamp` member variable :attr:`.Timestamp.epoch`
            is incremented immediately before :attr:`.Event.EPOCH_END`.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def epoch_checkpoint(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EPOCH_CHECKPOINT` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def predict_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.PREDICT_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def predict_batch_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.PREDICT_BATCH_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def predict_before_forward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.PREDICT_BATCH_FORWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def predict_after_forward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.PREDICT_AFTER_FORWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def predict_batch_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.PREDICT_BATCH_END` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def predict_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.PREDICT_END` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def eval_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EVAL_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def eval_batch_start(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EVAL_BATCH_START` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def eval_before_forward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EVAL_BATCH_FORWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def eval_after_forward(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EVAL_AFTER_FORWARD` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def eval_batch_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EVAL_BATCH_END` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def eval_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.EVAL_END` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def fit_end(self, state: State, logger: Logger) -> None:
        """Called on the :attr:`.Event.FIT_END` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        del state, logger  # unused
        pass

    def close(self, state: State, logger: Logger) -> None:
        """Called whenever the trainer finishes training, even when there is an exception.

        It should be used for clean up tasks such as flushing I/O streams and/or
        closing any files that may have been opened during the :attr:`.Event.INIT` event.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
        """
        pass

    def post_close(self) -> None:
        """Called after :meth:`close` has been invoked for each callback.

        Very few callbacks should need to implement :meth:`post_close`.
        This callback can be used to back up any data that may have
        been written by other callbacks during :meth:`close`.
        """
        pass



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The EfficientNet model family is a set of convolutional neural networks that can be used as the basis for a variety
of vision tasks, but were initially designed for image classification. The model family was designed to reach the
highest accuracy for a given computation budget during inference by simultaneously scaling model depth, model width, and
image resolution according to an empirically determined scaling law.

See the :doc:`Model Card </model_cards/efficientnet>` for more details.
"""
from composer.models.efficientnetb0.model import composer_efficientnetb0 as composer_efficientnetb0

__all__ = ['composer_efficientnetb0']

_task = 'Image Classification'
_dataset = 'ImageNet'
_name = 'EfficientNet-B0'
_quality = '76.63'
_metric = 'Top-1 Accuracy'
_ttt = '21h 48m'
_hparams = 'efficientnetb0.yaml'



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import functools
import logging
import warnings
from typing import Optional, Sequence, Union

import numpy as np
import torch
from torch.optim import Optimizer

from composer.algorithms.blurpool.blurpool_layers import BlurConv2d, BlurMaxPool2d
from composer.algorithms.warnings import NoEffectWarning
from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.utils import module_surgery

log = logging.getLogger(__name__)


def apply_blurpool(model: torch.nn.Module,
                   replace_convs: bool = True,
                   replace_maxpools: bool = True,
                   blur_first: bool = True,
                   min_channels: int = 16,
                   optimizers: Optional[Union[Optimizer, Sequence[Optimizer]]] = None) -> None:
    """Add anti-aliasing filters to strided :class:`torch.nn.Conv2d` and/or :class:`torch.nn.MaxPool2d` modules.

    These filters increase invariance to small spatial shifts in the input
    (`Zhang 2019 <http://proceedings.mlr.press/v97/zhang19a.html>`_).

    Args:
        model (:class:`torch.nn.Module`): the model to modify in-place
        replace_convs (bool, optional): replace strided :class:`torch.nn.Conv2d` modules with
            :class:`.BlurConv2d` modules. Default: ``True``.
        replace_maxpools (bool, optional): replace eligible :class:`torch.nn.MaxPool2d` modules
            with :class:`.BlurMaxPool2d` modules. Default: ``True``.
        blur_first (bool, optional): for ``replace_convs``, blur input before the associated
            convolution. When set to ``False``, the convolution is applied with
            a stride of 1 before the blurring, resulting in significant
            overhead (though more closely matching
            `the paper <http://proceedings.mlr.press/v97/zhang19a.html>`_).
            See :class:`.BlurConv2d` for further discussion. Default: ``True``.
        min_channels (int, optional): Skip replacing layers with in_channels < min_channels.
            Commonly used to prevent the blurring of the first layer. Default: 16.
        optimizers (torch.optim.Optimizer | Sequence[torch.optim.Optimizer], optional):
            Existing optimizers bound to ``model.parameters()``. All optimizers that have already been
            constructed with ``model.parameters()`` must be specified here so
            they will optimize the correct parameters.

            If the optimizer(s) are constructed *after* calling this function,
            then it is safe to omit this parameter. These optimizers will see
            the correct model parameters.

    Example:
        .. testcode::

            import composer.functional as cf
            from torchvision import models
            model = models.resnet50()
            cf.apply_blurpool(model)
    """
    transforms = {}
    if replace_maxpools:
        transforms[torch.nn.MaxPool2d] = BlurMaxPool2d.from_maxpool2d
    if replace_convs:
        transforms[torch.nn.Conv2d] = functools.partial(
            _maybe_replace_strided_conv2d,
            blur_first=blur_first,
            min_channels=min_channels,
        )
    module_surgery.replace_module_classes(model, optimizers=optimizers, policies=transforms)
    _log_surgery_result(model)


class BlurPool(Algorithm):
    """`BlurPool <http://proceedings.mlr.press/v97/zhang19a.html>`_ adds anti-aliasing filters to convolutional layers.

    This algorithm increases accuracy and invariance to small shifts in the input. It runs on
    :attr:`.Event.INIT`.

    Args:
        replace_convs (bool): replace strided :class:`torch.nn.Conv2d` modules with
            :class:`.BlurConv2d` modules. Default: ``True``.
        replace_maxpools (bool): replace eligible :class:`torch.nn.MaxPool2d` modules
            with :class:`.BlurMaxPool2d` modules. Default: ``True``.
        blur_first (bool): when ``replace_convs`` is ``True``, blur input before the
            associated convolution. When set to ``False``, the convolution is
            applied with a stride of 1 before the blurring, resulting in
            significant overhead (though more closely matching the paper).
            See :class:`.BlurConv2d` for further discussion. Default: ``True``.
        min_channels (int, optional): Skip replacing layers with in_channels < min_channels.
            Commonly used to prevent the blurring of the first layer. Default: 16.
    """

    def __init__(
        self,
        replace_convs: bool = True,
        replace_maxpools: bool = True,
        blur_first: bool = True,
        min_channels: int = 16,
    ) -> None:
        self.replace_convs = replace_convs
        self.replace_maxpools = replace_maxpools
        self.blur_first = blur_first
        self.min_channels = min_channels

        if self.replace_maxpools is False and self.replace_convs is False:
            raise ValueError(
                'Both replace_maxpool and replace_convs are set to False. BlurPool will not be modifying the model.')

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(replace_convs={self.replace_convs},replace_maxpools={self.replace_maxpools},blur_first={self.blur_first},min_channels={self.min_channels})'

    @staticmethod
    def required_on_load() -> bool:
        return True

    def match(self, event: Event, state: State) -> bool:
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        assert state.model is not None

        apply_blurpool(state.model,
                       optimizers=state.optimizers,
                       replace_convs=self.replace_convs,
                       replace_maxpools=self.replace_maxpools,
                       blur_first=self.blur_first,
                       min_channels=self.min_channels)
        self._log_results(event, state, logger)

    def _log_results(self, event: Event, state: State, logger: Logger) -> None:
        """Logs the result of BlurPool application, including the number of layers that have been replaced."""
        assert state.model is not None

        num_blurpool_layers = module_surgery.count_module_instances(state.model, BlurMaxPool2d)
        num_blurconv_layers = module_surgery.count_module_instances(state.model, BlurConv2d)

        # python logger
        log.info(f'Applied BlurPool to model {state.model.__class__.__name__} '
                 f'with replace_maxpools={self.replace_maxpools}, '
                 f'replace_convs={self.replace_convs}. '
                 f'Model now has {num_blurpool_layers} BlurMaxPool2d '
                 f'and {num_blurconv_layers} BlurConv2D layers.')

        logger.log_hyperparameters({
            'blurpool/num_blurpool_layers': num_blurpool_layers,
            'blurpool/num_blurconv_layers': num_blurconv_layers,
        })


def _log_surgery_result(model: torch.nn.Module):
    num_blurpool_layers = module_surgery.count_module_instances(model, BlurMaxPool2d)
    num_blurconv_layers = module_surgery.count_module_instances(model, BlurConv2d)
    if num_blurconv_layers == 0 and num_blurpool_layers == 0:
        warnings.warn(
            NoEffectWarning('Applying BlurPool did not change any layers. '
                            'No strided Conv2d or Pool2d layers were found.'))
    log.info(f'Applied BlurPool to model {model.__class__.__name__}. '
             f'Model now has {num_blurpool_layers} BlurMaxPool2d '
             f'and {num_blurconv_layers} BlurConv2D layers.')


def _maybe_replace_strided_conv2d(
    module: torch.nn.Conv2d,
    module_index: int,
    blur_first: bool,
    min_channels: int = 16,
):
    already_blurpooled = hasattr(module, '_already_blurpooled') and module._already_blurpooled
    if np.max(module.stride) > 1 and module.in_channels >= min_channels and not already_blurpooled:
        return BlurConv2d.from_conv2d(module, module_index, blur_first=blur_first)
    return None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Implements stochastic depth (`Huang et al, 2016 <https://arxiv.org/abs/1603.09382>`_) for ResNet blocks.

See :class:`.StochasticDepth`, the sample-wise stochastic depth :doc:`method card
</method_cards/stochastic_depth_samplewise>`, or the block-wise stochastic depth :doc:`method card
</method_cards/stochastic_depth>` for details.
"""

from composer.algorithms.stochastic_depth.stochastic_depth import StochasticDepth as StochasticDepth
from composer.algorithms.stochastic_depth.stochastic_depth import apply_stochastic_depth as apply_stochastic_depth

__all__ = ['StochasticDepth', 'apply_stochastic_depth']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Performance profiling tools.

The profiler gathers performance metrics during a training run that can be used to diagnose bottlenecks and
facilitate model development.

The metrics gathered include:

* Duration of each :class:`.Event` during training
* Time taken by the data loader to return a batch
* Host metrics such as CPU, system memory, disk and network utilization over time
* Execution order, latency and attributes of PyTorch operators and GPU kernels (see :doc:`torch:profiler`)

See the :doc:`Profiling Guide </trainer/performance_tutorials/profiling>` for additional information.
"""
from composer.profiler.json_trace_handler import JSONTraceHandler
from composer.profiler.marker import Marker
from composer.profiler.profiler import Profiler
from composer.profiler.profiler_action import ProfilerAction
from composer.profiler.profiler_schedule import cyclic_schedule
from composer.profiler.system_profiler import SystemProfiler
from composer.profiler.torch_profiler import TorchProfiler
from composer.profiler.trace_handler import TraceHandler

# All needs to be defined properly for sphinx autosummary
__all__ = [
    'Marker',
    'Profiler',
    'ProfilerAction',
    'TraceHandler',
    'cyclic_schedule',
    'JSONTraceHandler',
    'SystemProfiler',
    'TorchProfiler',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core gradient clipping classes and functions."""

from __future__ import annotations

import logging
from typing import Iterable, Optional, Union

import torch
from packaging import version

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.models import ComposerModel

log = logging.getLogger(__name__)

__all__ = ['GradientClipping', 'apply_gradient_clipping']


def apply_gradient_clipping(model: Union[ComposerModel, torch.nn.Module], clipping_type: str, clipping_threshold: float,
                            fsdp_enabled: bool):
    """Clips all gradients in model based on specified clipping_type.

    Args:
        model (ComposerModel or torch.nn.Module): The model that we want to apply gradient clipping.
        clipping_type ('adaptive', 'norm', 'value'): String denoting which type of
            gradient clipping to do. The options are: 'norm', which clips the gradient norm
            and uses `torch.nn.utils.clip_grad_norm_`, 'value', which clips gradient at
            a specified value and uses `torch.nn.utils.clip_grad_value_`, and 'adaptive',
            which clips all gradients based on gradient norm:parameter norm ratio using
            composer.algorithms.gradient_clipping.gradient_clipping._apply_agc.
        clipping_threshold (float, optional): Specifies what value to clip the gradients
            to (for 'value'), what values to clip the gradient norms to (for 'norm'), and
            threshold by which if grad_norm / weight_norm is greater than this threshold then
            scale gradients by this threshold * (weight_norm / grad_norm) (for 'adaptive').
        fsdp_enabled (bool): Bool of if the model is a FSDP model or not.
    """
    if fsdp_enabled:
        if version.parse(torch.__version__) < version.parse('1.13.0'):
            raise RuntimeError('To use FSDP with Composer, you must use torch>=1.13.0.')
        from torch.distributed.fsdp import FullyShardedDataParallel
        for module in model.modules():
            if isinstance(module, FullyShardedDataParallel):
                # We can only call grad clip on the parent instance, so we iterate through all
                # modules and try grad clipping and FSDP will throw an exception if we
                # clip any gradients that aren't a parent module
                try:
                    if clipping_type == 'norm':
                        module.clip_grad_norm_(max_norm=clipping_threshold)
                    elif clipping_type == 'value':
                        module.clip_grad_norm_(max_norm=clipping_threshold, norm_type=float('inf'))
                    else:
                        raise ValueError(f"clipping type must be 'norm' or 'value' with FSDP not {clipping_type}")
                except AssertionError as e:
                    # Catches the error message from PyTorch
                    if 'clip_grad_norm should only be called on the root (parent) instance' == str(e):
                        continue
                    else:
                        raise
        return
    parameters = model.parameters()
    if clipping_type == 'adaptive':
        _apply_agc(parameters, clipping_threshold=clipping_threshold)
    elif clipping_type == 'norm':
        torch.nn.utils.clip_grad_norm_(parameters, max_norm=clipping_threshold)
    elif clipping_type == 'value':
        torch.nn.utils.clip_grad_value_(parameters, clip_value=clipping_threshold)
    else:
        raise ValueError(f"clipping_type must be 'adaptive', 'norm', or 'value' not {clipping_type} ")


def _apply_agc(
    parameters: Union[torch.Tensor, Iterable[torch.Tensor]],
    clipping_threshold: float,
) -> None:
    """Clips all gradients in model based on ratio of gradient norms to parameter norms.
    Args:
        parameters (torch.Tensor or Iterable[torch.Tensor]): The parameters to of the
            model for whose gradients we will clip
        clipping_threshold (float, optional): The largest acceptable ratio between grad
            norms and parameter norms before clipping is done.
    """
    for param in parameters:
        if param.grad is None:
            continue

        # Detach weights and gradients, so the clipping operation is not added to
        # computational graph.
        weights = param.detach()
        grad = param.grad.detach()

        # Get clipped version of gradients.
        clipped_grad_coeff = _get_clipped_gradient_coeff(weights, grad, clipping_threshold=clipping_threshold)

        # Copy clipped gradients into param.grad attribute, so they can be accessed by
        # optimizer.
        grad.mul_(clipped_grad_coeff)


class GradientClipping(Algorithm):
    """Clips all gradients in model based on specified clipping_type.

    Runs on ``Event.AFTER_TRAIN_BATCH``.

    Example:
         .. testcode::

            from composer.algorithms import GradientClipping
            from composer.trainer import Trainer
            gc = GradientClipping(clipping_type='norm', clipping_threshold=0.1)
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[gc],
                optimizers=[optimizer]
            )

    Args:
        clipping_type ('adaptive', 'norm', 'value'): String denoting which type of
            gradient clipping to do. The options are: 'norm', which clips the gradient norm
            and uses `torch.nn.utils.clip_grad_norm_`, 'value', which clips gradient at
            a specified value and uses `torch.nn.utils.clip_grad_value_`, and 'adaptive',
            which clips all gradients based on gradient norm:parameter norm ratio using
            composer.algorithms.gradient_clipping.gradient_clipping._apply_agc.
        clipping_threshold (float, optional): Specifies what value to clip the gradients
            to (for 'value'), what values to clip the gradient norms to (for 'norm'), and
            threshold by which if grad_norm / weight_norm is greater than this threshold then
            scale gradients by this threshold * (weight_norm / grad_norm) (for 'adaptive').

    Raises:
        NotImplementedError: if deepspeed is enabled and clipping_type is not 'norm'.
        ValueError: if deepspeed is enabled and clipping_type is not 'norm'.
    """

    def __init__(self, clipping_type: str, clipping_threshold: float):
        self.clipping_type = clipping_type
        self.clipping_threshold = clipping_threshold

    def match(self, event: Event, state: State) -> bool:
        return event in [Event.INIT, Event.AFTER_TRAIN_BATCH]

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        if event == Event.INIT and state.deepspeed_config is not None:
            if self.clipping_type == 'norm':
                if self.clipping_threshold > 0:
                    state.deepspeed_config['gradient_clipping'] = self.clipping_threshold
                else:
                    raise ValueError(
                        f'Deepspeed only supports gradient clipping thresholds that are greater than zero, but the provided one is {self.clipping_threshold}'
                    )
            else:
                raise NotImplementedError(
                    f"Deepspeed only supports gradient clipping of type 'norm' not of type '{self.clipping_type}'")

        if event == Event.AFTER_TRAIN_BATCH and not state.deepspeed_enabled:
            apply_gradient_clipping(model=state.model,
                                    clipping_type=self.clipping_type,
                                    clipping_threshold=self.clipping_threshold,
                                    fsdp_enabled=state.fsdp_enabled)


def _get_clipped_gradient_coeff(weights: torch.Tensor, grad: torch.Tensor, clipping_threshold: float = 0.01):
    """Clips all gradients in model based on ratio of gradient norms to parameter norms.

    Gradients whose norms exceed

    .. math:: weight_norm * clipping_threshold

    are scaled down by

    .. math:: (weight_norm / grad_norm) * clipping_threshold.

    Args:
        weights (torch.Tensor): Tensor of weights (parameters) from the model.
        grad (torch.Tensor): Tensor of gradients of the loss with respect to the weights.
        clipping_threshold (float, optional): The largest acceptable ratio between grad
            norms and parameter norms before clipping is done.

    Return:
        clipped_grad_coeff (torch.Tensor): Coefficient of same shape as grad_norm equal to
            (weight_norm / grad_norm) * clipping_threshold for gradients whose norms
            that exceed weight_norm * clipping_threshold and one otherwise.
    """

    # Compute and clamp grad and weight norms.
    w_norm = _unitwise_norm(weights)
    grad_norm = _unitwise_norm(grad)

    # Gradients whose norms are greater than weight_norm * clipping_threhsold are
    # scaled down by (weight_norm * clipping_threhsold) / grad_norm.
    max_norm = w_norm.mul_(clipping_threshold)
    clipped_grad_coeff = max_norm.div_(grad_norm).nan_to_num_(nan=1.0).clamp_(max=1.0)

    return clipped_grad_coeff


def _unitwise_norm(tensor: torch.Tensor):
    """Implements unitwise norm as described in Brock et al, 2021.

    For 0D scalars of shape [], we trivially normalize with dim=0 which essentially returns the absolute value of the scalar.
    For 1D *.bias weights of shape [out_features], we normalize across entire vector -> dim=0.
    For 2D torch.nn.Linear weights of shape [out_features, in_features]: we normalize across in_features -> dim = 1
    For 4D torch.nn.Conv2d weights [out_channels, in_channels, kernel_height, kernel_width]:
        we normalize across [in_channels, kernel_height, kernel_width] -> dim = (1, 2, 3).
    If a 3D parameter were somehow in your model, we would normalize buy the last two dimensions -> dim = (1,2).

    Args:
        tensor (torch.Tensor): A parameter or gradient of the model.

    Returns:
        The appropriate L2 norm of the parameter or gradient as described above.
    """
    # 0D for scalars, 1D for bias vectors.
    if tensor.ndim <= 1:
        dim = 0
        keepdim = False
    # 2D corresponds to MLPs and 4D corresponds to ConvNets.
    else:
        dim = tuple(range(1, tensor.ndim))
        keepdim = True
    # L2 Norm.
    return torch.linalg.vector_norm(tensor, ord=2, dim=dim, keepdim=keepdim)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import subprocess
import sys
import textwrap
from pathlib import Path
from typing import List
from unittest.mock import Mock

import pytest

import composer
from composer.core import Engine, Event
from composer.core.algorithm import Algorithm
from composer.core.callback import Callback
from composer.core.state import State
from composer.loggers import Logger, LoggerDestination
from tests.common.events import EventCounterCallback


@pytest.fixture
def always_match_algorithms():
    return [
        Mock(**{
            'match.return.value': True,
            'apply.return_value': n,  # return encodes order
            'interpolate_loss': False,
        }) for n in range(5)
    ]


@pytest.fixture()
def dummy_logger(dummy_state: State):
    return Logger(dummy_state)


@pytest.fixture
def never_match_algorithms():
    attrs = {'match.return_value': False}
    return [Mock(**attrs) for _ in range(5)]


def run_event(event: Event, state: State, logger: Logger):
    runner = Engine(state, logger)
    return runner.run_event(event)


class DummyCallback(Callback):

    def __init__(self, file_path):
        self.file_path = file_path

    def init(self, state: State, logger: Logger):
        with open(self.file_path, 'a') as f:
            f.write('init callback, ')

    def batch_end(self, state: State, logger: Logger):
        with open(self.file_path, 'a') as f:
            f.write('on_batch_end callback, ')


class DummyLoggerDestination(LoggerDestination):

    def __init__(self, file_path):
        self.file_path = file_path

    def init(self, state: State, logger: Logger):
        with open(self.file_path, 'a') as f:
            f.write('init logger, ')

    def batch_end(self, state: State, logger: Logger):
        with open(self.file_path, 'a') as f:
            f.write('on_batch_end logger, ')


def test_engine_runs_callbacks_in_correct_order(dummy_state, tmp_path):
    file_path = tmp_path / Path('event_check.txt')
    dummy_state.callbacks = [DummyCallback(file_path), DummyLoggerDestination(file_path)]
    logger = Logger(dummy_state)
    engine = Engine(dummy_state, logger)
    engine.run_event(Event.INIT)
    engine.run_event(Event.BATCH_END)
    engine.run_event(Event.EPOCH_END)
    engine.close()
    expected_lines = ['init logger, init callback, on_batch_end callback, on_batch_end logger, ']
    with open(file_path, 'r') as f:
        actual_lines = f.readlines()
    assert expected_lines == actual_lines


@pytest.mark.parametrize('event', list(Event))
class TestAlgorithms:

    def test_algorithms_always_called(self, event: Event, dummy_state: State, always_match_algorithms: List[Algorithm],
                                      dummy_logger: Logger):
        dummy_state.algorithms = always_match_algorithms
        _ = run_event(event, dummy_state, dummy_logger)
        for algo in always_match_algorithms:
            algo.apply.assert_called_once()
            algo.match.assert_called_once()

    def test_algorithms_never_called(self, event: Event, dummy_state: State, never_match_algorithms: List[Algorithm],
                                     dummy_logger: Logger):
        dummy_state.algorithms = never_match_algorithms
        _ = run_event(event, dummy_state, dummy_logger)
        for algo in never_match_algorithms:
            algo.apply.assert_not_called()
            algo.match.assert_called_once()

    def test_engine_trace_all(self, event: Event, dummy_state: State, always_match_algorithms: List[Algorithm],
                              dummy_logger: Logger):
        dummy_state.algorithms = always_match_algorithms
        trace = run_event(event, dummy_state, dummy_logger)

        assert all([tr.run for tr in trace.values()])

    def test_engine_trace_never(self, event: Event, dummy_state: State, never_match_algorithms: List[Algorithm],
                                dummy_logger: Logger):
        dummy_state.algorithms = never_match_algorithms
        trace = run_event(event, dummy_state, dummy_logger)

        assert all([tr.run is False for tr in trace.values()])


def test_engine_is_dead_after_close(dummy_state: State, dummy_logger: Logger):
    # Create the trainer and run an event
    engine = Engine(dummy_state, dummy_logger)
    engine.run_event(Event.INIT)

    # Close it
    engine.close()

    # Assert it complains if you try to run another event
    with pytest.raises(RuntimeError):
        engine.run_event(Event.FIT_START)


class IsClosedCallback(Callback):

    def __init__(self) -> None:
        self.is_closed = True

    def init(self, state: State, logger: Logger) -> None:
        assert self.is_closed
        self.is_closed = False

    def close(self, state: State, logger: Logger) -> None:
        self.is_closed = True


def test_engine_closes_on_del(dummy_state: State, dummy_logger: Logger):
    # Create the trainer and run an event
    is_closed_callback = IsClosedCallback()
    dummy_state.callbacks.append(is_closed_callback)
    engine = Engine(dummy_state, dummy_logger)
    engine.run_event(Event.INIT)

    # Assert that there is just 2 -- once above, and once as the arg temp reference
    assert sys.getrefcount(engine) == 2

    # Implicitly close the engine
    del engine

    # Assert it is closed
    assert is_closed_callback.is_closed


class DummyTrainer:
    """Helper to simulate what the trainer does w.r.t. events"""

    def __init__(self, state: State, logger: Logger) -> None:
        self.engine = Engine(state, logger)
        self.engine.run_event(Event.INIT)

    def close(self):
        self.engine.close()


def test_engine_triggers_close_only_once(dummy_state: State, dummy_logger: Logger):
    # Create the trainer and run an event
    is_closed_callback = IsClosedCallback()
    dummy_state.callbacks.append(is_closed_callback)

    # Create the trainer
    trainer = DummyTrainer(dummy_state, dummy_logger)

    # Close the trainer
    trainer.close()

    # Assert it is closed
    assert is_closed_callback.is_closed

    # Create a new trainer with the same callback. Should implicitly trigger __del__ AFTER
    # AFTER DummyTrainer was constructed
    trainer = DummyTrainer(dummy_state, dummy_logger)

    # Assert it is open
    assert not is_closed_callback.is_closed


def test_engine_errors_if_previous_trainer_was_not_closed(dummy_state: State, dummy_logger: Logger):
    # Create the trainer and run an event
    is_closed_callback = IsClosedCallback()
    dummy_state.callbacks.append(is_closed_callback)

    # Create the trainer
    _ = DummyTrainer(dummy_state, dummy_logger)

    # Assert the callback is open
    assert not is_closed_callback.is_closed

    # Create a new trainer with the same callback. Should raise an exception
    # because trainer.close() was not called before
    with pytest.raises(RuntimeError,
                       match=r'Cannot create a new trainer with an open callback or logger from a previous trainer'):
        DummyTrainer(dummy_state, dummy_logger)


def check_output(proc: subprocess.CompletedProcess):
    # Check the subprocess output, and raise an exception with the stdout/stderr dump if there was a non-zero exit
    # The `check=True` flag available in `subprocess.run` does not print stdout/stderr
    if proc.returncode == 0:
        return
    error_msg = textwrap.dedent(f"""\
        Command {proc.args} failed with exit code {proc.returncode}.
        ----Begin stdout----
        {proc.stdout}
        ----End stdout------
        ----Begin stderr----
        {proc.stderr}
        ----End stderr------""")

    raise RuntimeError(error_msg)


@pytest.mark.parametrize('exception', [True, False])
def test_engine_closes_on_atexit(exception: bool):
    # Running this test via a subprocess, as atexit() must trigger

    code = textwrap.dedent("""\
    from composer import Trainer, Callback
    from tests.common import SimpleModel

    class CallbackWithConditionalCloseImport(Callback):
        def post_close(self):
            import requests

    model = SimpleModel(3, 10)
    cb = CallbackWithConditionalCloseImport()
    trainer = Trainer(
        model=model,
        callbacks=[cb],
        max_duration="1ep",
        train_dataloader=None,
    )
    """)
    if exception:
        # Should raise an exception, since no dataloader was provided
        code += 'trainer.fit()'

    git_root_dir = os.path.join(os.path.dirname(composer.__file__), '..')
    proc = subprocess.run(['python', '-c', code], cwd=git_root_dir, text=True, capture_output=True)
    if exception:
        # manually validate that there was no a conditional import exception
        assert 'ImportError: sys.meta_path is None, Python is likely shutting down' not in proc.stderr
    else:
        check_output(proc)


def test_logging(
    caplog: pytest.LogCaptureFixture,
    dummy_state: State,
    dummy_logger: Logger,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test that engine logs statements as expected"""
    caplog.set_level(logging.DEBUG, logger=Engine.__module__)
    # Include a callback, since most logging happens around callback events
    dummy_state.callbacks = [EventCounterCallback()]

    monkeypatch.setenv('ENGINE_DEBUG', '1')
    engine = Engine(dummy_state, dummy_logger)
    engine.run_event('INIT')
    engine.close()

    # Validate that we have the expected log entries
    assert caplog.record_tuples == [
        ('composer.core.engine', 10, '[ep=0][ba=0][event=INIT]: Running event'),
        ('composer.core.engine', 10, '[ep=0][ba=0][event=INIT]: Running callback EventCounterCallback'),
        ('composer.core.engine', 10, 'Closing the engine'),
        ('composer.core.engine', 10, 'Closing callback EventCounterCallback'),
        ('composer.core.engine', 10, 'Post-closing callback EventCounterCallback'),
    ]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from composer.core import MemoryFormat
from composer.datasets.synthetic import SyntheticBatchPairDataset
from composer.utils import dist


def build_mnist_dataloader(
    datadir: str,
    global_batch_size: int,
    is_train: bool = True,
    download: bool = True,
    drop_last: bool = True,
    shuffle: bool = True,
    **dataloader_kwargs: Any,
) -> DataLoader:
    """Builds an MNIST dataloader.

    Args:
        datadir (str): Path to the data directory
        global_batch_size (int): Global batch size.
        is_train (bool): Whether to load the training data or validation data. Default:
            ``True``.
        download (bool, optional): Whether to download the dataset, if needed. Default:
            ``True``.
        drop_last (bool): Drop remainder samples. Default: ``True``.
        shuffle (bool): Shuffle the dataset. Default: ``True``.
        **dataloader_kwargs (Any): Additional settings for the dataloader (e.g. num_workers, etc.)
    """
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()

    transform = transforms.Compose([transforms.ToTensor()])

    with dist.run_local_rank_zero_first():
        dataset = datasets.MNIST(
            datadir,
            train=is_train,
            download=dist.get_local_rank() == 0 and download,
            transform=transform,
        )

    sampler = dist.get_sampler(dataset, drop_last=drop_last, shuffle=shuffle)

    return DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler,
        drop_last=drop_last,
        **dataloader_kwargs,
    )


def build_synthetic_mnist_dataloader(
    global_batch_size: int,
    is_train: bool = True,
    drop_last: bool = True,
    shuffle: bool = True,
    num_unique_samples: int = 100,
    device: str = 'cpu',
    memory_format: MemoryFormat = MemoryFormat.CONTIGUOUS_FORMAT,
    **dataloader_kwargs: Any,
) -> DataLoader:
    """Builds a synthetic MNIST dataset.

    Args:
        global_batch_size (int): Global batch size.
        is_train (bool): Whether to load the training data or validation data. Default:
            ``True``.
        drop_last (bool): Drop remainder samples. Default: ``True``.
        shuffle (bool): Shuffle the dataset. Default: ``True``.
        num_unique_samples (int): number of unique samples in synthetic dataset. Default: ``100``.
        device (str): device with which to load the dataset. Default: ``cpu``.
        memory_format (:class:`composer.core.MemoryFormat`): memory format of the tensors. Default: ``CONTIGUOUS_FORMAT``.
        **dataloader_kwargs (Any): Additional settings for the dataloader (e.g. num_workers, etc.)
    """
    if global_batch_size % dist.get_world_size() != 0:
        raise ValueError(
            f'global_batch_size ({global_batch_size}) must be divisible by world_size ({dist.get_world_size()}).')
    batch_size = global_batch_size // dist.get_world_size()
    dataset = SyntheticBatchPairDataset(
        total_dataset_size=60_000 if is_train else 10_000,
        data_shape=[1, 28, 28],
        num_classes=10,
        num_unique_samples_to_create=num_unique_samples,
        device=device,
        memory_format=memory_format,
    )
    sampler = dist.get_sampler(dataset, drop_last=drop_last, shuffle=shuffle)

    return DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler,
        drop_last=drop_last,
        **dataloader_kwargs,
    )



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helper utilities for configuring deterministic training to ensure reproducibility.

.. note::

    For deterministic model initialization, :func:`~.seed_all` and/or
    :func:`~.configure_deterministic_mode` should be
    invoked before creating and initializing a model, before creating the :class:`~.Trainer`.
    For example:

    .. testsetup::

        import functools
        import torch.nn
        import warnings

        warnings.filterwarnings(action="ignore", message="Deterministic mode is activated.")

        MyModel = Model

    .. doctest::

        >>> import torch.nn
        >>> from composer.utils import reproducibility
        >>> reproducibility.configure_deterministic_mode()
        >>> reproducibility.seed_all(42)
        >>> model = MyModel()
        >>> def init_weights(m):
        ...     if isinstance(m, torch.nn.Linear):
        ...         torch.nn.init.xavier_uniform(m.weight)
        >>> # model will now be deterministically initialized, since the seed is set.
        >>> init_weights(model)
        >>> trainer = Trainer(model=model, seed=42)

    Note that the seed must also be passed to the Trainer, otherwise the Trainer
    would generate a random seed based on the timestamp (see :func:`~.get_random_seed`).

    .. testcleanup::

        warnings.resetwarnings()

Attributes:
    MAX_SEED (int): The maximum allowed seed, which is :math:`2^{32} - 1`.
"""

from __future__ import annotations

import logging
import os
import random
import textwrap
import time
import warnings
from typing import Any, Dict, List

import numpy as np
import torch
import torch.backends.cudnn

from composer.utils import dist

__all__ = [
    'configure_deterministic_mode',
    'get_random_seed',
    'seed_all',
    'get_rng_state',
    'load_rng_state',
    'MAX_SEED',
]

log = logging.getLogger(__name__)

# seeds must be 32-bit unsigned integers
MAX_SEED = 2**32 - 1


def configure_deterministic_mode():
    """Configure PyTorch deterministic mode.

    .. note::

        When using the :class:`.Trainer`, you can use the ``deterministic_mode`` flag
        instead of invoking this function directly.
        For example:

        .. testsetup::

            import warnings

            warnings.filterwarnings(action="ignore", message="Deterministic mode is activated.")

        .. doctest::

            >>> trainer = Trainer(deterministic_mode=True)

        .. testcleanup::

            warnings.resetwarnings()

        However, to configure deterministic mode for operations before the trainer is initialized, manually invoke this
        function at the beginning of your training script.

    .. note::

        When training on a GPU, this function must be invoked before any CUDA operations.

    .. note::

        Deterministic mode degrades performance. Do not use outside of testing and debugging.
    """
    torch.use_deterministic_algorithms(True)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    # See https://pytorch.org/docs/stable/generated/torch.use_deterministic_algorithms.html
    # and https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
    warnings.warn('Deterministic mode is activated. This will negatively impact performance.', category=UserWarning)


def get_random_seed() -> int:
    """Get a randomly created seed to use for seeding rng objects.

    .. warning::

        This random seed is NOT cryptographically secure.

    Returns:
        int: A random seed.
    """
    rng = random.Random(int(time.time_ns()))  # get a new RNG does not respect the current seed
    seed = rng.randint(0, MAX_SEED)
    assert seed >= 0 and seed <= MAX_SEED, 'seed should be on this range'
    return seed


def seed_all(seed: int):
    """Seed all rng objects.

    .. note::

        When using the :class:`.Trainer`, you can use the ``seed`` parameter
        instead of invoking this function directly.
        For example:

        .. doctest::

            >>> trainer = Trainer(seed=42)

        However, to configure the random seed for operations before the trainer is initialized, manually invoke this
        function at the beginning of your training script.

    Args:
        seed (int): The random seed
    """
    if seed < 0 or seed > MAX_SEED:
        raise ValueError(f'Seed {seed} is invalid. It must be on [0; 2^32 - 1]')
    log.info('Setting seed to %d', seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    # torch.manual_seed may call manual_seed_all but calling it again here
    # to make sure it gets called at least once
    torch.cuda.manual_seed_all(seed)


def get_rng_state() -> List[Dict[str, Any]]:
    """The state of the RNG objects.

    Returns:
        List[Dict[str, Any]]: A list of RNG State Dicts, indexed by global rank.
    """
    rng_state = {
        'python': random.getstate(),
        'numpy': np.random.get_state(),
        'torch': torch.random.get_rng_state(),
    }
    if torch.cuda.is_available() and torch.cuda.is_initialized():
        # This will not be compatible with model parallelism
        rng_state['cuda'] = torch.cuda.get_rng_state()

    return dist.all_gather_object(rng_state)


def load_rng_state(rng_state_dicts: List[Dict[str, Any]]):
    """Restore the RNG state.

    Args:
        rng_state_dicts (List[Dict[str, Any]]): The list of RNG state dicts to restore,
            as returned by :func:`get_rng_state`.
    """
    if dist.get_world_size() > len(rng_state_dicts):
        warnings.warn(
            textwrap.dedent(f"""\
                The current world size ({dist.get_world_size()} is greater than the number of RNG state(s) serialized
                ({len(rng_state_dicts)}). Only the first {len(rng_state_dicts)} rank(s) will have their RNG restored.
                """))
    if dist.get_world_size() < len(rng_state_dicts):
        warnings.warn(
            textwrap.dedent(f"""\
            The current world size ({dist.get_world_size()} is less than the number of RNG state(s) serialized
            ({len(rng_state_dicts)}). Only the first {dist.get_world_size()} RNG state(s) will be consumed;
            the remaining will be ignored."""))

    if dist.get_global_rank() < len(rng_state_dicts):
        rng_state_dict = rng_state_dicts[dist.get_global_rank()]
        torch.set_rng_state(rng_state_dict['torch'])
        random.setstate(rng_state_dict['python'])
        np.random.set_state(rng_state_dict['numpy'])

        is_cuda_available = torch.cuda.is_available() and torch.cuda.is_initialized()
        has_cuda_rng_state = 'cuda' in rng_state_dict

        log.debug('Restoring the RNG state')

        if is_cuda_available and has_cuda_rng_state:
            torch.cuda.set_rng_state(rng_state_dict['cuda'])

        if is_cuda_available and not has_cuda_rng_state:
            warnings.warn(
                textwrap.dedent(f"""\
                The checkpoint did not include the CUDA RNG state. The CUDA RNG will have a
                non-deterministic state."""))
        if not is_cuda_available and has_cuda_rng_state:
            warnings.warn(
                textwrap.dedent(f"""\
                The checkpoint included CUDA RNG state, but CUDA is not being used.
                As such, the CUDA RNG state will be ignored."""))



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Utility for uploading to and downloading from cloud object stores."""
import io
import os
import pathlib
import uuid
from typing import Any, Callable, Dict, Optional, Union

from requests.exceptions import ConnectionError
from urllib3.exceptions import ProtocolError

from composer.utils.import_helpers import MissingConditionalImportError
from composer.utils.iter_helpers import iterate_with_callback
from composer.utils.object_store.object_store import ObjectStore, ObjectStoreTransientError

__all__ = ['LibcloudObjectStore']


class LibcloudObjectStore(ObjectStore):
    """Utility for uploading to and downloading from object (blob) stores, such as Amazon S3.

    .. rubric:: Example

    Here's an example for an Amazon S3 bucket named ``MY_CONTAINER``:

    >>> from composer.utils import LibcloudObjectStore
    >>> object_store = LibcloudObjectStore(
    ...     provider="s3",
    ...     container="MY_CONTAINER",
    ...     provider_kwargs={
    ...         "key": "AKIA...",
    ...         "secret": "*********",
    ...     }
    ... )
    >>> object_store
    <composer.utils.object_store.libcloud_object_store.LibcloudObjectStore object at ...>

    Args:
        provider (str): Cloud provider to use. Valid options are:

            * :mod:`~libcloud.storage.drivers.atmos`
            * :mod:`~libcloud.storage.drivers.auroraobjects`
            * :mod:`~libcloud.storage.drivers.azure_blobs`
            * :mod:`~libcloud.storage.drivers.backblaze_b2`
            * :mod:`~libcloud.storage.drivers.cloudfiles`
            * :mod:`~libcloud.storage.drivers.digitalocean_spaces`
            * :mod:`~libcloud.storage.drivers.google_storage`
            * :mod:`~libcloud.storage.drivers.ktucloud`
            * :mod:`~libcloud.storage.drivers.local`
            * :mod:`~libcloud.storage.drivers.minio`
            * :mod:`~libcloud.storage.drivers.nimbus`
            * :mod:`~libcloud.storage.drivers.ninefold`
            * :mod:`~libcloud.storage.drivers.oss`
            * :mod:`~libcloud.storage.drivers.rgw`
            * :mod:`~libcloud.storage.drivers.s3`

            .. seealso:: :doc:`Full list of libcloud providers <libcloud:storage/supported_providers>`

        container (str): The name of the container (i.e. bucket) to use.
        provider_kwargs (Dict[str, Any], optional):  Keyword arguments to pass into the constructor
            for the specified provider. These arguments would usually include the cloud region
            and credentials.

            Common keys are:

            * ``key`` (str): API key or username to be used (required).
            * ``secret`` (str): Secret password to be used (required).
            * ``secure`` (bool): Whether to use HTTPS or HTTP. Note: Some providers only support HTTPS, and it is on by default.
            * ``host`` (str): Override hostname used for connections.
            * ``port`` (int): Override port used for connections.
            * ``api_version`` (str): Optional API version. Only used by drivers which support multiple API versions.
            * ``region`` (str): Optional driver region. Only used by drivers which support multiple regions.

            .. seealso:: :class:`libcloud.storage.base.StorageDriver`

        key_environ (str, optional): Environment variable name for the API Key. Only used
            if 'key' is not in ``provider_kwargs``. Default: None.
        secret_environ (str, optional): Envrionment varaible for the Secret password. Only
            used if 'secret' is not in ``provider_kwargs``. Default: None.
    """

    def __init__(self,
                 provider: str,
                 container: str,
                 chunk_size: int = 1_024 * 1_024,
                 key_environ: Optional[str] = None,
                 secret_environ: Optional[str] = None,
                 provider_kwargs: Optional[Dict[str, Any]] = None) -> None:
        try:
            from libcloud.storage.providers import get_driver
        except ImportError as e:
            raise MissingConditionalImportError('libcloud', 'apache-libcloud') from e
        provider_cls = get_driver(provider)
        if provider_kwargs is None:
            provider_kwargs = {}

        if 'key' not in provider_kwargs and \
           key_environ and key_environ in os.environ:
            provider_kwargs['key'] = os.environ[key_environ]

        if 'secret' not in provider_kwargs and \
           secret_environ and secret_environ in os.environ:
            provider_kwargs['secret'] = os.environ[secret_environ]

        self.chunk_size = chunk_size
        self._provider_name = provider
        self._provider = provider_cls(**provider_kwargs)
        self._container = self._provider.get_container(container)

    def get_uri(self, object_name: str):
        return f'{self._provider_name}://{self._container.name}/{object_name}'

    def upload_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        callback: Optional[Callable[[int, int], None]] = None,
    ):
        with open(filename, 'rb') as f:
            stream = iterate_with_callback(_file_to_iterator(f, self.chunk_size),
                                           os.fstat(f.fileno()).st_size, callback)
            try:
                self._provider.upload_object_via_stream(
                    stream,
                    container=self._container,
                    object_name=object_name,
                )
            except Exception as e:
                self._ensure_transient_errors_are_wrapped(e)

    def _ensure_transient_errors_are_wrapped(self, exc: Exception):
        from libcloud.common.types import LibcloudError
        if isinstance(exc, (LibcloudError, ProtocolError, TimeoutError, ConnectionError)):
            if isinstance(exc, LibcloudError):
                # The S3 driver does not encode the error code in an easy-to-parse manner
                # So first checking if the error code is non-transient
                is_transient_error = any(x in str(exc) for x in ('408', '409', '425', '429', '500', '503', '504'))
                if not is_transient_error:
                    raise exc
            raise ObjectStoreTransientError() from exc
        raise exc

    def _get_object(self, object_name: str):
        """Get object from object store.

        Args:
            object_name (str): The name of the object.
        """
        from libcloud.storage.types import ObjectDoesNotExistError
        try:
            return self._provider.get_object(self._container.name, object_name)
        except ObjectDoesNotExistError as e:
            raise FileNotFoundError(f'Object not found: {self.get_uri(object_name)}') from e
        except Exception as e:
            self._ensure_transient_errors_are_wrapped(e)

    def get_object_size(self, object_name: str) -> int:
        return self._get_object(object_name).size

    def download_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        overwrite: bool = False,
        callback: Optional[Callable[[int, int], None]] = None,
    ):
        if os.path.exists(filename) and not overwrite:
            # If the file already exits, short-circuit and skip the download
            raise FileExistsError(f'filename {filename} exists and overwrite was set to False.')

        obj = self._get_object(object_name)
        # Download first to a tempfile, and then rename, in case if the file gets corrupted in transit
        tmp_filepath = str(filename) + f'.{uuid.uuid4()}.tmp'
        try:
            with open(tmp_filepath, 'wb+') as f:
                stream = self._provider.download_object_as_stream(obj, chunk_size=self.chunk_size)
                for chunk in iterate_with_callback(stream, obj.size, callback):
                    f.write(chunk)
        except Exception as e:
            # The download failed for some reason. Make a best-effort attempt to remove the temporary file.
            try:
                os.remove(tmp_filepath)
            except OSError:
                pass
            self._ensure_transient_errors_are_wrapped(e)

        # The download was successful.
        if overwrite:
            os.replace(tmp_filepath, filename)
        else:
            os.rename(tmp_filepath, filename)


def _file_to_iterator(f: io.IOBase, chunk_size: int):
    while True:
        byte = f.read(chunk_size)
        if byte == b'':
            break
        yield byte



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import copy
import os

import pytest
import torch
from torch.utils.data import DataLoader
from torchmetrics.classification import MulticlassAccuracy

from composer.algorithms import GatedLinearUnits
from composer.loggers import RemoteUploaderDownloader
from composer.metrics.nlp import LanguageCrossEntropy, MaskedAccuracy
from composer.models import HuggingFaceModel
from composer.trainer import Trainer
from composer.utils import dist, get_device, inference, reproducibility
from tests.common import device
from tests.common.datasets import RandomTextClassificationDataset, RandomTextLMDataset
from tests.common.models import SimpleTransformerClassifier, SimpleTransformerMaskedLM


def get_model_embeddings(model):
    if isinstance(model, HuggingFaceModel):
        return model.model.bert.embeddings.word_embeddings.weight
    elif isinstance(model, SimpleTransformerClassifier) or isinstance(model, SimpleTransformerMaskedLM):
        return model.transformer_base.embedding.weight
    else:
        raise ValueError('Unsure how to get embeddings layer from model.')


def pretraining_test_helper(tokenizer, model, algorithms, tmp_path, device):
    transformers = pytest.importorskip('transformers')

    pretraining_model_copy = copy.deepcopy(model)
    pretraining_train_dataset = RandomTextLMDataset(size=8,
                                                    vocab_size=tokenizer.vocab_size,
                                                    sequence_length=4,
                                                    use_keys=True)

    collator = transformers.DataCollatorForLanguageModeling(tokenizer=tokenizer, mlm_probability=0.15)
    pretraining_train_dataloader = DataLoader(pretraining_train_dataset,
                                              batch_size=4,
                                              sampler=dist.get_sampler(pretraining_train_dataset),
                                              collate_fn=collator)
    pretraining_eval_dataloader = DataLoader(pretraining_train_dataset,
                                             batch_size=4,
                                             sampler=dist.get_sampler(pretraining_train_dataset),
                                             collate_fn=collator)

    pretraining_trainer = Trainer(model=pretraining_model_copy,
                                  train_dataloader=pretraining_train_dataloader,
                                  save_folder=str(tmp_path / 'pretraining_checkpoints'),
                                  max_duration='1ep',
                                  seed=17,
                                  algorithms=algorithms,
                                  device=device)
    pretraining_trainer.fit()
    reproducibility.seed_all(17)  # seed so that the masking is the same
    pretraining_trainer.eval(pretraining_eval_dataloader)

    loaded_pretraining_trainer = Trainer(model=model,
                                         load_path=str(tmp_path / 'pretraining_checkpoints' / 'latest-rank0.pt'),
                                         seed=17,
                                         algorithms=algorithms,
                                         device=device)

    reproducibility.seed_all(17)  # seed so that the masking is the same
    loaded_pretraining_trainer.eval(pretraining_eval_dataloader)

    original_ce = pretraining_trainer.state.eval_metrics['eval']['LanguageCrossEntropy']
    loaded_ce = loaded_pretraining_trainer.state.eval_metrics['eval']['LanguageCrossEntropy']
    assert original_ce.compute() > 0.0
    assert original_ce.compute() == loaded_ce.compute()

    return str(tmp_path / 'pretraining_checkpoints' / 'latest-rank0.pt')


def finetuning_test_helper(tokenizer, model, algorithms, checkpoint_path, pretraining_model, tmp_path, device):
    finetuning_model_copy = copy.deepcopy(model)

    finetuning_train_dataset = RandomTextClassificationDataset(size=8,
                                                               vocab_size=tokenizer.vocab_size,
                                                               sequence_length=4,
                                                               num_classes=3,
                                                               use_keys=isinstance(model, HuggingFaceModel))
    finetuning_train_dataloader = DataLoader(finetuning_train_dataset,
                                             batch_size=4,
                                             sampler=dist.get_sampler(finetuning_train_dataset))
    finetuning_eval_dataloader = DataLoader(finetuning_train_dataset,
                                            batch_size=4,
                                            sampler=dist.get_sampler(finetuning_train_dataset))

    remote_dir = str(tmp_path / 'object_store')
    os.makedirs(remote_dir, exist_ok=True)

    rud = RemoteUploaderDownloader(
        bucket_uri='libcloud://.',
        backend_kwargs={
            'provider': 'local',
            'container': '.',
            'provider_kwargs': {
                'key': remote_dir,
            },
        },
        num_concurrent_uploads=1,
        use_procs=False,
        upload_staging_folder=str(tmp_path / 'staging_folder'),
    )

    finetuning_embedding_layer = get_model_embeddings(model)
    pretraining_embedding_layer = get_model_embeddings(pretraining_model)
    # The pretraining weights have not yet been loaded into the finetuning model
    assert not torch.equal(finetuning_embedding_layer.cpu(), pretraining_embedding_layer.cpu())
    finetuning_trainer = Trainer(model=model,
                                 train_dataloader=finetuning_train_dataloader,
                                 save_folder='finetuning_checkpoints',
                                 load_path=checkpoint_path,
                                 load_weights_only=True,
                                 loggers=[rud],
                                 max_duration='1ep',
                                 seed=17,
                                 algorithms=algorithms,
                                 device=device)
    # Now they have been loaded
    assert torch.equal(finetuning_embedding_layer.cpu(), pretraining_embedding_layer.cpu())
    finetuning_trainer.fit()
    finetuning_trainer.eval(finetuning_eval_dataloader)

    loaded_finetuning_trainer = Trainer(model=finetuning_model_copy,
                                        load_path='finetuning_checkpoints/latest-rank0.pt',
                                        load_object_store=rud,
                                        seed=17,
                                        algorithms=algorithms,
                                        device=device)

    loaded_finetuning_trainer.eval(finetuning_eval_dataloader)

    original_acc = finetuning_trainer.state.eval_metrics['eval']['MulticlassAccuracy']
    loaded_acc = loaded_finetuning_trainer.state.eval_metrics['eval']['MulticlassAccuracy']
    assert original_acc.compute() > 0.0
    assert original_acc.compute() == loaded_acc.compute()

    return loaded_finetuning_trainer, finetuning_eval_dataloader, rud, 'finetuning_checkpoints/latest-rank0.pt'


def inference_test_helper(finetuning_output_path, rud, finetuning_model, algorithms, original_input, original_output,
                          tmp_path, save_format, device):
    inference_trainer = Trainer(model=finetuning_model,
                                load_path=finetuning_output_path,
                                load_weights_only=True,
                                loggers=[rud],
                                seed=17,
                                algorithms=algorithms,
                                device=device)

    os.mkdir(tmp_path / 'inference_checkpoints')
    sample_input = (original_input, {})

    inference.export_for_inference(model=inference_trainer.state.model,
                                   save_format=save_format,
                                   save_path=str(tmp_path / 'inference_checkpoints' / f'exported_model.{save_format}'),
                                   sample_input=sample_input)

    copied_batch = copy.deepcopy(original_input)

    if save_format == 'onnx':
        onnx = pytest.importorskip('onnx')
        ort = pytest.importorskip('onnxruntime')
        loaded_inference_model = onnx.load(str(tmp_path / 'inference_checkpoints' / 'exported_model.onnx'))
        onnx.checker.check_model(loaded_inference_model)
        ort_session = ort.InferenceSession(str(tmp_path / 'inference_checkpoints' / 'exported_model.onnx'))

        for key, value in copied_batch.items():
            copied_batch[key] = value.numpy()
        loaded_model_out = ort_session.run(None, copied_batch)
    elif save_format == 'torchscript':
        loaded_inference_model = torch.jit.load(str(tmp_path / 'inference_checkpoints' / 'exported_model.torchscript'))
        loaded_inference_model.eval()
        loaded_model_out = loaded_inference_model(copied_batch)
    else:
        raise ValueError('Unsupported save format')

    torch.testing.assert_close(
        loaded_model_out[1] if isinstance(loaded_model_out, list) else loaded_model_out.detach().numpy(),
        original_output.detach().numpy()
        if isinstance(original_output, torch.Tensor) else original_output.logits.detach().numpy())


@device('cpu', 'gpu')
# Note: the specificity of these settings are due to incompatibilities (e.g. the simpletransformer model is not traceable)
@pytest.mark.parametrize('model_type,algorithms,save_format', [('tinybert_hf', [GatedLinearUnits], 'onnx'),
                                                               ('simpletransformer', [], 'torchscript')])
def test_full_nlp_pipeline(model_type, algorithms, save_format, tiny_bert_tokenizer, tmp_path, request, device):
    """This test is intended to exercise our full pipeline for NLP.

    To this end, it performs pretraining, loads the pretrained model with a classification head for finetuning
    and finetunes it, exports the model for inference, and loads it back in to make predictions.
    """
    pytest.importorskip('libcloud')
    pytest.importorskip('transformers')

    algorithms = [algorithm() for algorithm in algorithms]

    device = get_device(device)

    tiny_bert_model = None
    if model_type == 'tinybert_hf':
        tiny_bert_model = request.getfixturevalue('tiny_bert_model')

    # pretraining
    if model_type == 'tinybert_hf':
        assert tiny_bert_model is not None
        pretraining_metrics = [LanguageCrossEntropy(ignore_index=-100), MaskedAccuracy(ignore_index=-100)]
        pretraining_model = HuggingFaceModel(tiny_bert_model,
                                             tiny_bert_tokenizer,
                                             use_logits=True,
                                             metrics=pretraining_metrics)
    elif model_type == 'simpletransformer':
        pretraining_model = SimpleTransformerMaskedLM(vocab_size=tiny_bert_tokenizer.vocab_size)
    else:
        raise ValueError('Unsupported model type')
    pretraining_output_path = pretraining_test_helper(tiny_bert_tokenizer, pretraining_model, algorithms, tmp_path,
                                                      device)

    # finetuning
    if model_type == 'tinybert_hf':
        finetuning_metric = MulticlassAccuracy(num_classes=3, average='micro')
        hf_finetuning_model, _ = HuggingFaceModel.hf_from_composer_checkpoint(
            pretraining_output_path,
            model_instantiation_class='transformers.AutoModelForSequenceClassification',
            model_config_kwargs={'num_labels': 3})
        finetuning_model = HuggingFaceModel(model=hf_finetuning_model,
                                            tokenizer=tiny_bert_tokenizer,
                                            use_logits=True,
                                            metrics=[finetuning_metric])
    elif model_type == 'simpletransformer':
        finetuning_model = SimpleTransformerClassifier(vocab_size=tiny_bert_tokenizer.vocab_size, num_classes=3)
    else:
        raise ValueError('Unsupported model type.')

    finetuning_model_copy = copy.deepcopy(finetuning_model)
    finetuning_trainer, finetuning_dataloader, rud, finetuning_output_path = finetuning_test_helper(
        tiny_bert_tokenizer, finetuning_model, algorithms, pretraining_output_path, pretraining_model, tmp_path, device)

    # inference
    batch = next(iter(finetuning_dataloader))
    finetuning_trainer.state.model.to('cpu')
    finetuning_trainer.state.model.eval()
    original_output = finetuning_trainer.state.model(batch)
    inference_test_helper(finetuning_output_path, rud, finetuning_model_copy, algorithms, batch, original_output,
                          tmp_path, save_format, device)
